import numpy as np
from .poincare import PoincareHamiltonian
from sympy import symbols, S, binomial, summation, sqrt, cos, sin, Function,atan2,expand_trig,diff,Matrix
from sympy import lambdify as sym_lambdify
from sympy.functions import elliptic_f,elliptic_k
from sympy.core import pi
from .disturbing_function import  _p1_p2_from_k_nu, evaluate_df_coefficient_delta_expansion
from .disturbing_function import df_coefficient_C,evaluate_df_coefficient_dict,get_df_coefficient_symbol
from .disturbing_function import  list_resonance_terms, _nucombos, _lcombos
from .poincare import get_re_im_components, _get_Lambda0_symbol, _get_a0_symbol
from .hamiltonian import _my_elliptic_e, _lambdify_kwargs
import warnings

class FirstOrderGeneratingFunction(PoincareHamiltonian):
    """
    A class representing a generating function that maps from the 'osculating'
    canonical variables of the full :math:`N`-body problem to 'mean' variables
    of an 'averaged' Hamiltonian or normal form.  

    The generating function is constructed to first order in planet-star mass
    ratios by specifying indivdual cosine terms to be eliminated from the full
    Hamiltonian.

    This class is a sub-class of :class:`celmech.poincare.PoincareHamiltonian`
    and disturbing function terms to be eliminated are added in the same manner
    that disturbing function terms can be added to
    :class:`celmech.poincare.PoincareHamiltonian`.

    Attributes
    ----------
    chi : sympy expression
        Symbolic expression for the generating function.
    N_chi : sympy expression
        Symbolic expression for the generating function with numerical values
        of paramteters substituted where applicable.
    state : :class:`celmech.poincare.Poincare`
        A set of Poincare variables to which transformations are applied.
    N : int
        Number of particles
    particles : list
        List of :class:`~celmech.poincare.PoincareParticle` objects making up
        the system.
    """
    def __init__(self,pvars):
        super(FirstOrderGeneratingFunction,self).__init__(pvars)
        self.H = S(0)

    @property
    def chi(self):
        return self.H
    @property
    def N_chi(self):
        return self.N_H

    def _get_approximate_corrections(self,y):
        return self.flow_func(*y).reshape(-1)
        #corrections = np.zeros(y.shape)
        #for i,deriv_func in enumerate(self.N_derivs):
        #    corrections[i] = deriv_func(*y)
        #return corrections

    def osculating_to_mean_state_vector(self,y,approximate=True,**integrator_kwargs):
        r"""
        Convert a state vector of canonical variables of the 
        the un-averaged :math:`N`-body Hamiltonian to a state
        vector of mean variables used by a normal form.

        Arguments
        ---------
        y : array-like
          State vector of un-averaged variables.

        approximate : bool, optional
          If True, Lie transformation is computed approximately
          to first order.  In other words, the approximation

          .. math::
            \exp[{\cal L}_{\chi}]f \approx f + \left[f,\chi \right]

        is used. If False, the exact Lie transformation is computed
        numerically.

        Returns
        -------
        ymean : array-like
          State vector of transformed (averaged) variables.
        """
        if approximate:
            yarr = np.atleast_1d(y)
            corrections = self._get_approximate_corrections(yarr)
            return yarr - corrections
        else:
            self.integrator.set_initial_value(y,t=0)
            self.integrator.integrate(-1,**integrator_kwargs)
            return self.integrator.y

    def mean_to_osculating_state_vector(self,y,approximate=True,**integrator_kwargs):
        r"""
        Convert a state vector of canonical variables of mean 
        variables used by a normal form to un-averaged variables
        of the full :math:`N`-body Hamiltonian.

        Arguments
        ---------
        y : array-like
          State vector of 'averaged' canonical variables.

        approximate : bool, optional
            If True, Lie transformation is computed approximately
            to first order.  In other words, the approximation

            .. math::
              \exp[{\cal L}_{\chi}]f \approx f + \left[f,\chi \right]

            is used. Otherwise, the exact Lie transformation is computed
            numerically.

        Returns
        -------
        yosc : array-like
          State vector of osculating canonical variables.
        """
        if approximate:
            yarr = np.atleast_1d(y)
            corrections = self._get_approximate_corrections(yarr)
            return yarr + corrections
        else:
            self.integrator.set_initial_value(y,t=0)
            self.integrator.integrate(+1,**integrator_kwargs)
            return self.integrator.y

    def osculating_to_mean(self,**integrator_kwargs):
        """
        Convert the :attr:`state <celmech.generating_functions.FirstOrderGeneratingFunction.state>`'s
        variables from osculating
        to mean canonical variables.
        """
        y_osc = self.state.values
        y_mean = self.osculating_to_mean_state_vector(y_osc)
        self.state.values = y_mean

    def mean_to_osculating(self,**integrator_kwargs):
        """
        Convert the :attr:`state <celmech.generating_functions.FirstOrderGeneratingFunction.state>`'s
        variables from osculating
        to mean canonical variables.
        """
        y_mean = self.state.values
        y_osc = self.mean_to_osculating_state_vector(y_mean)
        self.state.values = y_osc

    def add_zeroth_order_term(self,indexIn=1,indexOut=2):
        r"""
        Add generating function term that elimiates 
        planet-planet interactions to 0th order in 
        inclinations and eccentricities.
        
        The added generating function term cancels the term

        .. math:: 
            -\frac{Gm_im_j}{a_j}\left(\frac{1}{\sqrt{1+\alpha^2-2\cos(\lambda_j-\lambda_i)}} - \frac{1}{2}b_{1/2}^{(0)}(\alpha) -\frac{\cos(\lambda_j-\lambda_i)}{\sqrt{\alpha}} \right)

        from the Hamiltonian to first order in planet-star mass ratios.

        Arguments
        ---------
        indexIn : int, optional
          Index of inner planet
        indexOut : int, optional
          Index of outer planet
        """
        G = symbols('G')
        mIn,muIn,MIn,LambdaIn,lambdaIn = symbols('m{0},mu{0},M{0},Lambda{0},lambda{0}'.format(indexIn)) 
        mOut,muOut,MOut,LambdaOut,lambdaOut = symbols('m{0},mu{0},M{0},Lambda{0},lambda{0}'.format(indexOut)) 
        aOut_inv = G*MOut*muOut*muOut / LambdaOut / LambdaOut  
        prefactor = -G * mIn * mOut * aOut_inv
        psi = lambdaOut - lambdaIn
        aIn = (LambdaIn/muIn)**2 / G / MIn
        aOut = (LambdaOut/muOut)**2 / G / MOut
        alpha = aIn/aOut
        omega_syn = self.kvec_to_omega((1,-1,0,0,0,0),indexIn,indexOut)
        m = alpha*alpha
        om_alpha = 1-alpha
        om_alpha_sq = om_alpha * om_alpha
        F = elliptic_f(psi/2,-4 * alpha / om_alpha_sq) / om_alpha
        psi_integral = 2 * F - 2 * psi * elliptic_k(m) / pi - sin(psi) / sqrt(alpha)
        term = prefactor * psi_integral / omega_syn
        self.H += term

    def get_mean_motion(self,index):
        G,mu,M,Lambda = symbols('G,mu{0},M{0},Lambda{0}'.format(index)) 
        return G*G*M*M*mu*mu*mu / Lambda / Lambda / Lambda

    def kvec_to_omega(self,kvec,indexIn,indexOut):
        nIn = self.get_mean_motion(indexIn)
        nOut = self.get_mean_motion(indexOut)
        return kvec[0] * nOut + kvec[1] * nIn

    def add_cosine_term(self,k_vec,max_order=None,l_max=0,nu_vecs=None,l_vecs=None,indexIn=1,indexOut=2,eccentricities=True,inclinations=True):
        """
        Add a specific cosine term to the disturbing function up to an optional max_order in the eccentricities and inclinations.

        Arguments
        ---------
        k_vec : Tuple of 6 ints 
            Coefficients (k1, k2, k3, k4, k5, k6) for a cosine term of the form 
            cos(k1*lambdaOut + k2*lambdaIn + k3*pomegaIn + k4*pomegaOut + k5*OmegaIn + k6*OmegaOut)
            Note the ks must sum to zero, and by convention we write lambdaOut first, e.g., (3, -2, -1, 0, 0, 0) is
            cos(3*lambdaOut - 2*lambdaIn - pomega1)
        max_order : int, optional
            Maximum order to go to in the (combined) eccentricities and inclinations.
            Can specify either max_order OR nu_vecs (see below), but not both (most users will use max_order).
            If neither are passed, max_order defaults to the leading order of the cosine term specified by k_vec.
        l_max : int, optional
            Maximum degree of expansion in :math:`\delta = (\Lambda-\Lambda_0)/\Lambda_0` to include in cosine coefficients. Default is 0.
            Can specify either l_max OR l_vecs (see below), but not both (most users will use l_max).
        nu_vecs: List of 4-tuples, optional
            A list of the specific combinations of nu indices to include for the cosine term coefficient, e.g., [(0, 0, 0, 0), (1, 0, 0, 0), ...]
            See paper and examples for definition and use.
            Can specify either max_order OR nu_vecs, but not both (max_order makes more sense for most use cases).
        l_vecs: List of 2-tuples, optional
            A list of the specific combinations of l indices to include for the cosine term coefficient, e.g., [(0, 0), (1, 0), (2, 0), ...] 
            See paper and examples for definition and use.
            Can specify either l_max OR l_vecs, but not both (l_max makes more sense for most use cases).
            One use case for passing particular l_vecs is if one body is massless, so the other Lambda doesn't vary (and doesn't need expansion)
        indexIn : int, optional
            Index of inner planet. Default 1.
        indexOut : int, optional
            Index of outer planet. Default 2.
        eccentricities: bool, optional
            By default, includes all eccentricity terms.
            Can set to False to exclude any eccentricity terms (e.g., fixed circular orbits).
        inclinations: bool, optional
            By default, includes all inclination terms.
            Can set to False to exclude any inclination terms (e.g., co-planar systems).
        """
        if np.sum(k_vec) != 0:
            raise AttributeError("Invalid k_vec={0}. The coefficients must sum to zero to satisfy the d'Alembert relation.".format(k_vec))
        k1,k2,k3,k4,k5,k6 = k_vec
        k_vec = tuple(k_vec) # ensure k_vec is a tuple
        if nu_vecs:
            nu_vecs = [tuple(nu_vec) for nu_vec in nu_vecs]
            if max_order:
                raise AttributeError('Must pass EITHER max_order OR nu_vecs to add_cos_term, but not both. See docstring.')
        else:
            # this is the leading order for the cosine term chosen (k3+k4+k5+k6 = -(k1+k2) by d'Alembert)
            min_order = abs(k3) + abs(k4) + abs(k5) + abs(k6)
            if not max_order:
                max_order = min_order
            nu_max = (max_order - min_order)//2 # each nu contributes 2 powers of e or i, so divide by 2 rounding down
            nu_vecs = []
            for nu_tot in range(nu_max+1):
                nu_vecs += _nucombos(nutot=nu_tot) # Make list of [(nu1, nu2, nu3, nu4), ... ] tuples
        if l_vecs:
            l_vecs = [tuple(l_vec) for l_vec in l_vecs]
            if l_max > 0:
                raise AttributeError('Can only pass l_max OR l_vecs to add_cos_term. See docstring.')
        else:
            l_vecs = []
            for l_tot in range(l_max+1):
                l_vecs += _lcombos(ltot=l_tot) # Make list of [(l1, l2), ... ] tuples
      
        try:
            if np.array(l_vecs).shape[1] != 2:
                raise
        except:
            raise AttributeError("l_vecs = {0} must be a list of (l1, l2) pairs, e.g., [(0, 0), (0, 1), ...]".format(l_vecs))

        try:
            if np.array(nu_vecs).shape[1] != 4:
                raise
        except:
            raise AttributeError("nu_vecs = {0} must be a list of (nu1, nu2, nu3, nu4) tuples, e.g., [(0, 0, 0, 0), (1, 0, 0, 0), ...]".format(nu_vecs))
        lmax = np.max([l1+l2 for l1, l2 in l_vecs]) # maximum total l=l1+l2

        G = symbols('G')
        mIn,muIn,MIn,LambdaIn,lambdaIn,kappaIn,etaIn,sigmaIn,rhoIn = symbols('m{0},mu{0},M{0},Lambda{0},lambda{0},kappa{0},eta{0},sigma{0},rho{0}'.format(indexIn)) 
        mOut,muOut,MOut,LambdaOut,lambdaOut,kappaOut,etaOut,sigmaOut,rhoOut = symbols('m{0},mu{0},M{0},Lambda{0},lambda{0},kappa{0},eta{0},sigma{0},rho{0}'.format(indexOut)) 
        
        Lambda0In,Lambda0Out = _get_Lambda0_symbol(indexIn),_get_Lambda0_symbol(indexOut)
        alpha_sym = symbols(r"\alpha_{{{0}\,{1}}}".format(indexIn,indexOut))
        alpha_val = self.H_params[alpha_sym]
        aOut0 = _get_a0_symbol(indexOut)
        deltaIn = (LambdaIn - Lambda0In) / Lambda0In
        deltaOut = (LambdaOut - Lambda0Out) / Lambda0Out
        # alpha = aIn/aOut
        # Resonance components
        
        omega = self.kvec_to_omega(k_vec,indexIn,indexOut)
        omega_inv = 1/omega
        for nu_vec in nu_vecs:
            if eccentricities == False and k_nu_depend_on_eccentricities(k_vec, nu_vec) == True:
                continue
            if inclinations == False and k_nu_depend_on_inclinations(k_vec, nu_vec) == True:
                continue

            C_dict = df_coefficient_C(*k_vec, *nu_vec)#k1,k2,k3,k4,k5,k6,nu1,nu2,nu3,nu4)
            p1,p2 = _p1_p2_from_k_nu(k_vec,nu_vec)
            C_delta_expansion_dict = evaluate_df_coefficient_delta_expansion(C_dict,p1,p2,lmax,alpha_val)
            Ctot = 0
            for l_vec,C_val in C_delta_expansion_dict.items():
                if l_vec not in l_vecs: # have to calculate all terms up to lmax, but only consider if in l_vecs
                    continue
                if l_vecs and (indexIn,indexOut,(k_vec,nu_vec,l_vec)) in self.resonance_indices:
                    warnings.warn("Cosine term k_vec={0}, nu_vec={1}, l_vec={2} already included Hamiltonian; no new term added.".format(k_vec, nu_vec, l_vec))
                    continue
                if l_vecs and (indexIn,indexOut,(tuple([-k for k in k_vec]),nu_vec,l_vec)) in self.resonance_indices:
                    warnings.warn("Cosine term k_vec={0}, nu_vec={1}, l_vec={2} already included Hamiltonian; no new term added.".format(k_vec, nu_vec, l_vec))
                    continue
                self.resonance_indices.append((indexIn,indexOut,(k_vec,nu_vec,l_vec)))
                Csym = get_df_coefficient_symbol(*k_vec,*nu_vec,*l_vec,indexIn,indexOut)
                self.H_params[Csym] = C_val
                l1,l2=l_vec
                Ctot += Csym * deltaIn**l1 * deltaOut**l2
            nu1,nu2,nu3,nu4 = nu_vec
            rtLIn = sqrt(Lambda0In)
            rtLOut = sqrt(Lambda0Out)
            xin,yin = get_re_im_components(kappaIn/rtLIn ,-etaIn / rtLIn,k3)
            xout,yout = get_re_im_components( kappaOut/rtLOut, -etaOut/rtLOut,k4)
            uin,vin = get_re_im_components(sigmaIn/rtLIn/2, -rhoIn/rtLIn/2,k5)
            uout,vout = get_re_im_components(sigmaOut/rtLOut/2, -rhoOut/rtLOut/2,k6)

            re = uin*uout*xin*xout - vin*vout*xin*xout - uout*vin*xout*yin - uin*vout*xout*yin - uout*vin*xin*yout - uin*vout*xin*yout - uin*uout*yin*yout + vin*vout*yin*yout
            im = uout*vin*xin*xout + uin*vout*xin*xout + uin*uout*xout*yin - vin*vout*xout*yin + uin*uout*xin*yout - vin*vout*xin*yout - uout*vin*yin*yout - uin*vout*yin*yout
            
            GammaIn = (kappaIn*kappaIn + etaIn*etaIn)/2
            GammaOut = (kappaOut*kappaOut + etaOut*etaOut)/2
            QIn = (sigmaIn*sigmaIn + rhoIn*rhoIn)/2
            QOut = (sigmaOut*sigmaOut + rhoOut*rhoOut)/2
            
            eIn_sq_term = (2 * GammaIn / Lambda0In )**nu3
            eOut_sq_term = (2 * GammaOut / Lambda0Out )**nu4
            incIn_sq_term = ( QIn / Lambda0In / 2 )**nu1
            incOut_sq_term = ( QOut / Lambda0Out / 2 )**nu2
            
            # Update internal Hamiltonian
            prefactor1 = -G * mIn * mOut / aOut0
            prefactor2 = eIn_sq_term * eOut_sq_term * incIn_sq_term * incOut_sq_term 
            trig_term = re * sin(k1 * lambdaOut + k2 * lambdaIn) + im * cos(k1 * lambdaOut + k2 * lambdaIn)
            self.H += prefactor1 * Ctot * prefactor2 * trig_term * omega_inv

    def get_perturbative_solution(self,expression,free_variables=[],correction_only=False,lambdify=False,time_symbol=None):
        r"""
        Calculate a solution for the time evolution of an expression using
        first-order perturbation theory.

        Arguments
        ---------
        expression : sympy expression
            Expression in terms of canonical variables for which 
            to compute a perturbative solution.
        free_variables : list, optional
            List of canonical variables to leave undetermined in the 
            derived expression. By default, `free_variables` is empty
            and numerical values are substituted for all canonical 
            variables
        correction_only : bool, optional
            If `True`, only return the perturbative correction to 
            the `expression`. I.e., if `expression` is 
            :math:`f(q,p)` then only return :math:`[f(q,p),\chi(q,p)]`.
            Otherwise, return :math:`f(p,q) + [f(q,p),\chi(q,p)].
            Default is `False`.
        lambdify : bool, optional
            If `True`, return a function using sympy.lambdify.
            The list of arguments accepted by the returned function
            will the list `free_variables`, followed by the 
            time at which to evaluate the expression.
        time_symbol : sympy.Symbol, optional
            Symbol to use for denoting time. If no symbol is given,
            :math:`t` is used.
            
        
        Returns
        -------
        result: sympy expression or function
            An expression for the solution as a function of time.
        """
        subsrule = {k:v for k,v in self.qp.items() if k not in free_variables}

        if time_symbol is None:
            t = symbols('t')
        else:
            t = time_symbol

        for i in range(1,self.N):
            n=self.get_mean_motion(i).subs(self.H_params)
            lsymb = symbols("lambda{}".format(i))
            exprn = n * t + lsymb
            subsrule[lsymb] = exprn.subs(subsrule)

        if correction_only:
            pt_solution = self.N_Lie_deriv(expression)
        else:
            pt_solution = expression + self.N_Lie_deriv(expression)
        result = pt_solution.subs(subsrule)
        if lambdify:
            args = list(free_variables) + [t]
            return sym_lambdify(args,result , **_lambdify_kwargs)
        else:
            return result



import numpy as np
import warnings
from abc import ABC, abstractmethod
from .hamiltonian import Hamiltonian
from .disturbing_function import DFCoeff_C,eval_DFCoeff_dict,get_DFCoeff_symbol
from .disturbing_function import terms_list_to_HamiltonianCoefficients_dict, _add_dicts, resonant_secular_contribution_dictionary,_consolidate_dictionary_terms
from .transformations import masses_to_jacobi, masses_from_jacobi
from .poincare import Poincare
from .miscellaneous import getOmegaMatrix
from scipy.linalg import solve as lin_solve
from scipy.linalg import expm
from numpy import sqrt

class EvolutionOperator(ABC):
    def __init__(self,initial_state,dt):
        self._dt = dt
        self.state = initial_state
        self.particles = initial_state.particles
    def _state_vector_to_individual_vectors(self,state_vec):
        return np.reshape(state_vec,(-1,6))

    @abstractmethod
    def apply(self):
        pass
    
    @abstractmethod
    def apply_to_state_vector(self,state_vec):
        pass

    @property
    @abstractmethod
    def dt(self):
        return self._dt
    @dt.setter
    def dt(self,value):
        self._dt = dt

class KeplerianEvolutionOperator(EvolutionOperator):
    def __init__(self,initial_state,dt):
        super(KeplerianEvolutionOperator,self).__init__(initial_state,dt)
        self.G =  self.state.G
        self.mu = np.array([p.mu for p in self.state.particles[1:]]) 
        self.M = np.array([p.M for p in self.state.particles[1:]]) 
        self.GGMMmmm = (self.G*self.M)**2 * self.mu**3

    def apply(self):
        lambda_dot = self.get_lambda_dot()
        dlambda = self.dt * lambda_dot
        ps = self.particles
        for p,dl in zip(ps[1:],dlambda):
            p.l += dl

    def apply_to_state_vector(self,state_vector):
        vecs = self._state_vector_to_individual_vectors(state_vector)
        L = vecs[:,2]
        lambda_dot = self.GGMMmmm/ L / L /L
        dlambda = self.dt * lambda_dot
        vecs[:,3] += dlambda
        return vecs.reshape(-1)

    
    def get_lambda_dot(self):
        ps = self.particles
        L = np.array([p.Lambda for p in ps[1:]])
        lambda_dot = self.GGMMmmm/L/L/L
        return lambda_dot

    @property
    def dt(self):
        return super().dt
    @dt.setter
    def dt(self,val):
        self._dt = val
    
class LinearInclinationResonancOperator(EvolutionOperator):
    r"""
    Evolution operator for linear equation of the form:
    .. math::
        dy/dt = -i e^{2i\theta} A \cdot y^* 
    Equations of motion derived from second-order resonances,
    expanded to leading order in inclinations can be written in this form.

    Parameters
    ----------
    A : ndarray, shape (M,M)
        The matrix appearing in the right-hand side
        of the equation.
    b : ndarray, shape (M)
        The vector in the r.h.s. of the equation.
    dt : float
        The timestep of the opertor.
    """
    def __init__(self,initial_state,indexIn,indexOut,res_vec, A,dt):
        super(LinearInclinationResonancOperator,self).__init__(initial_state,dt)        
        self.A = A
        self.Ainv = np.linalg.inv(A)
        vals,T = np.linalg.eigh(self.A)
        self.T = T
        self.Ttr = T.transpose()
        self.D = np.diag(vals)
        self.eigs = vals
        self._dt = dt
        self.cosh_lmbda_dt = np.cosh(self._dt * self.eigs)
        self.sinh_lmbda_dt = np.sinh(self._dt * self.eigs)
        self.indexIn = indexIn
        self.indexOut = indexOut
        self.particleIn = self.particles[indexIn]
        self.particleOut = self.particles[indexOut]
        self.res_vec = res_vec
        self.Lambdas_vec = np.array([-1 * res_vec[1] , res_vec[0]])
        self.Lambdas_Mtrx = np.linalg.inv([ self.Lambdas_vec , [1,1]])
    @property
    def indices(self):
        return self.indexIn-1,self.indexOut-1
    @property
    def dt(self):
        return super().dt
    @dt.setter
    def dt(self,val):
        self._dt = val
        self.cosh_lmbda_dt = np.cosh(val * self.eigs)
        self.sinh_lmbda_dt = np.sinh(val * self.eigs)
    def _get_rho_vec(self):
        return np.array([self.particleIn.rho,self.particleOut.rho])
    def _get_sigma_vec(self):
        return np.array([self.particleIn.sigma,self.particleOut.sigma])
    def _get_theta(self):
        lvec = np.array([ self.particleIn.l, self.particleOut.l])
        return self.res_vec @ lvec
    def _get_Lambdas(self):
        return np.array([self.particleIn.Lambda,self.particleOut.Lambda])
    def _get_Qs(self):
        return np.array([self.particleIn.Q,self.particleOut.Q])
    def _set_rho_vec(self,rho_vec):
        rhoIn,rhoOut = rho_vec
        self.particleIn.rho = rhoIn
        self.particleOut.rho = rhoOut
    def _set_sigma_vec(self,sigma_vec):
        sigmaIn,sigmaOut = sigma_vec
        self.particleIn.sigma = sigmaIn
        self.particleOut.sigma = sigmaOut
    def _get_RS_vecs(self,s,c):
        rho = self._get_rho_vec()
        sigma = self._get_sigma_vec()
        R = self.Ttr @ rho 
        S = self.Ttr @ sigma 
        return R,S
    def _RS_to_rhosigma(self,R,S):
        r = self.T @ R 
        s = self.T @ S
        return r,s
    def _RSmap(self,Rvec,Svec,s,c):
        s2 = 2 * s * c
        c2 = c*c - s*s
        S1vec = self.cosh_lmbda_dt * Svec  +  self.sinh_lmbda_dt * (s2 * Svec + c2 * Rvec)
        R1vec = self.cosh_lmbda_dt * Rvec  +  self.sinh_lmbda_dt * (c2 * Svec - s2 * Rvec)
        return R1vec,S1vec
    def _advance_inclination_variables(self):
        theta = self._get_theta()   
        s,c = np.sin(theta),np.cos(theta)
        R,S = self._get_RS_vecs(s,c)
        R1,S1  = self._RSmap(R,S,s,c)
        rho1,sigma1 = self._RS_to_rhosigma(R1,S1)
        self._set_rho_vec(rho1)
        self._set_sigma_vec(sigma1)

    def _set_Lambdas(self,Lambdas):
        self.particleIn.Lambda = Lambdas[0]
        self.particleOut.Lambda = Lambdas[1]

    def apply(self):
        Lambdas = self._get_Lambdas()
        Qs0  = self._get_Qs()
        C1 = self.Lambdas_vec @ Lambdas
        C2 = np.sum(Lambdas) - np.sum(Qs0)
        self._advance_inclination_variables()
        Qs1 = self._get_Qs()
        Lambdas1 = self.Lambdas_Mtrx @ np.array([ C1, C2 + np.sum(Qs1) ])
        self._set_Lambdas(Lambdas1)

    def apply_to_state_vector(self,state_vector):
        vecs = self._state_vector_to_individual_vectors(state_vector)
        Lambdas = vecs[self.indices,2]
        lambdas = vecs[self.indices,3]
        sigma,rho = vecs[self.indices,4],vecs[self.indices,5]
        Qs0 = 0.5 *  (rho**2 + sigma**2)
        C1 = self.Lambdas_vec @ Lambdas
        C2 = np.sum(Lambdas) - np.sum(Qs0)
        theta = self.res_vec @ lambdas
        s,c = np.sin(theta),np.cos(theta)
        R = self.Ttr @ rho
        S = self.Ttr @ sigma
        R1,S1 = self._RSmap(R,S,s,c)
        rho1,sigma1=self._RS_to_rhosigma(R1,S1)
        vecs[self.indices,4] = sigma1
        vecs[self.indices,5] = rho1
        Qs1 = 0.5 * (rho1**2 + sigma1**2)
        Lambdas1 = self.Lambdas_Mtrx @ np.array([ C1, C2 + np.sum(Qs1) ])
        vecs[self.indices,2] = Lambdas1
        return vecs.reshape(-1)
        
class LinearEccentricityResonancOperator(EvolutionOperator):
    r"""
    Evolution operator for linear equation of the form:
    .. math::
        dx/dt = -i e^{2i\theta} A \cdot x^* - \frac{i}{2}e^{i\theta} b
    Equations of motion derived from first- and second-order resonances,
    expanded to second order in eccentricities can be written in this form.

    An instance of `LinearEquationEvolutionOperator` provides
    a callable function that gives the evolution of h and k, where
        .. math::
            x = k - i h
    for a timestep dt and fixed angle value $\theta$.

    Parameters
    ----------
    A : ndarray, shape (M,M)
        The matrix appearing in the right-hand side
        of the equation.
    b : ndarray, shape (M)
        The vector in the r.h.s. of the equation.
    dt : float
        The timestep of the opertor.
    """
    def __init__(self,initial_state,indexIn,indexOut,res_vec, A, b,dt):
        super(LinearEccentricityResonancOperator,self).__init__(initial_state,dt)        
        self.A = A
        self.Ainv = np.linalg.inv(A)
        self.b = b
        self.Ainv_dot_b = self.Ainv @ b
        vals,T = np.linalg.eigh(self.A)
        self.T = T
        self.Ttr = T.transpose()
        self.D = np.diag(vals)
        self.eigs = vals
        self._dt = dt
        self.cosh_lmbda_dt = np.cosh(self._dt * self.eigs)
        self.sinh_lmbda_dt = np.sinh(self._dt * self.eigs)
        self.indexIn = indexIn
        self.indexOut = indexOut
        self.particleIn = self.particles[indexIn]
        self.particleOut = self.particles[indexOut]
        self.res_vec = res_vec
        self.Lambdas_vec = np.array([-1 * res_vec[1] , res_vec[0]])
        self.Lambdas_Mtrx = np.linalg.inv([ self.Lambdas_vec , [1,1]])
    @property
    def dt(self):
        return super().dt
    @property
    def indices(self):
        return self.indexIn-1,self.indexOut-1

    @dt.setter
    def dt(self,val):
        self._dt = val
        self.cosh_lmbda_dt = np.cosh(val * self.eigs)
        self.sinh_lmbda_dt = np.sinh(val * self.eigs)

    def _get_eta_vec(self):
        return np.array([self.particleIn.eta,self.particleOut.eta])

    def _get_kappa_vec(self):
        return np.array([self.particleIn.kappa,self.particleOut.kappa])

    def _get_theta(self):
        lvec = np.array([ self.particleIn.l, self.particleOut.l])
        return self.res_vec @ lvec

    def _get_Lambdas(self):
        return np.array([self.particleIn.Lambda,self.particleOut.Lambda])
    def _get_Gammas(self):
        return np.array([self.particleIn.Gamma,self.particleOut.Gamma])

    def _set_eta_vec(self,eta_vec):
        etaIn,etaOut = eta_vec
        self.particleIn.eta = etaIn
        self.particleOut.eta = etaOut
        
    def _set_kappa_vec(self,kappa_vec):
        kappaIn,kappaOut = kappa_vec
        self.particleIn.kappa = kappaIn
        self.particleOut.kappa = kappaOut

    def _get_HK_vecs(self,s,c):
        eta = self._get_eta_vec()
        kappa = self._get_kappa_vec()
        H = self.Ttr @ (eta - 0.5 * _rt2 * self.Ainv_dot_b * s)
        K = self.Ttr @ (kappa + 0.5 * _rt2 * self.Ainv_dot_b * c)
        return H,K

    def _HK_to_etakappa(self,H,K,s,c):
        h = self.T @ H + 0.5 * _rt2 * self.Ainv_dot_b * s
        k = self.T @ K - 0.5 * _rt2 * self.Ainv_dot_b * c
        return h,k

    def _HKmap(self,Hvec,Kvec,s,c):
        s2 = 2 * s * c
        c2 = c*c - s*s
        K1vec = self.cosh_lmbda_dt * Kvec  +  self.sinh_lmbda_dt * (s2 * Kvec + c2 * Hvec)
        H1vec = self.cosh_lmbda_dt * Hvec  +  self.sinh_lmbda_dt * (c2 * Kvec - s2 * Hvec)
        return H1vec,K1vec

    def _advance_eccentricity_variables(self):
        theta = self._get_theta()   
        s,c = np.sin(theta),np.cos(theta)
        H,K = self._get_HK_vecs(s,c)
        H1,K1  = self._HKmap(H,K,s,c)
        eta1,kappa1 = self._HK_to_etakappa(H1,K1,s,c)
        self._set_eta_vec(eta1)
        self._set_kappa_vec(kappa1)
    def _set_Lambdas(self,Lambdas):
        self.particleIn.Lambda = Lambdas[0]
        self.particleOut.Lambda = Lambdas[1]

    def apply(self):
        Lambdas = self._get_Lambdas()
        Gammas0  = self._get_Gammas()
        C1 = self.Lambdas_vec @ Lambdas
        C2 = np.sum(Lambdas) - np.sum(Gammas0)
        self._advance_eccentricity_variables()
        Gammas1 = self._get_Gammas()
        Lambdas1 = self.Lambdas_Mtrx @ np.array([ C1, C2 + np.sum(Gammas1) ])
        self._set_Lambdas(Lambdas1)

    def apply_to_state_vector(self,state_vector):
        vecs = self._state_vector_to_individual_vectors(state_vector)
        Lambdas = vecs[self.indices,2]
        lambdas = vecs[self.indices,3]
        theta = self.res_vec @ lambdas
        kappa,eta = vecs[self.indices,0],vecs[self.indices,1]
        Gammas0 = 0.5 *  (kappa**2 + eta**2)
        C1 = self.Lambdas_vec @ Lambdas
        C2 = np.sum(Lambdas) - np.sum(Gammas0)
        s,c = np.sin(theta),np.cos(theta)
        H = self.Ttr @ (eta - 0.5 * _rt2 * self.Ainv_dot_b * s)
        K = self.Ttr @ (kappa + 0.5 * _rt2 * self.Ainv_dot_b * c)
        H1,K1  = self._HKmap(H,K,s,c)
        eta1,kappa1 = self._HK_to_etakappa(H1,K1,s,c)
        vecs[self.indices,0] = kappa1
        vecs[self.indices,1] = eta1
        Gammas1 = 0.5 * (kappa1**2 + eta1**2)
        Lambdas1 = self.Lambdas_Mtrx @ np.array([ C1, C2 + np.sum(Gammas1) ])
        vecs[self.indices,2] = Lambdas1
        return vecs.reshape(-1)

class FirstOrderEccentricityResonanceOperator(LinearEccentricityResonancOperator):
    def __init__(self,initial_state,dt,j,indexIn=1,indexOut=2,Lambda0=None):
        res_vec = np.array([1-j,j])
        pIn = initial_state.particles[indexIn]
        pOut = initial_state.particles[indexOut]
        self.mIn = pIn.m
        self.MIn = pIn.M
        self.mOut = pOut.m
        self.MOut = pOut.M
        self.j = j
        if Lambda0 is None:
            _pvars = Poincare(initial_state.G,[pIn,pOut])
            _,Lambda0 = get_res_chain_reference_Lambdas_and_semimajor_axes(_pvars,[j-1])
            _,self.Lambda0In,self.Lambda0Out = Lambda0
        else:
            self.Lambda0In = Lambda0[indexIn]
            self.Lambda0Out = Lambda0[indexOut]
        Amtrx,bvec = get_first_order_eccentricity_resonance_matrix_and_vector(
                self.j,
                initial_state.G,
                self.mIn,
                self.mOut,
                self.MIn,
                self.MOut,
                self.Lambda0In,
                self.Lambda0Out
        )
        super(FirstOrderEccentricityResonanceOperator,self).__init__(initial_state,indexIn,indexOut,res_vec ,Amtrx,bvec,dt)

class SecondOrderEccentricityResonanceOperator(LinearEccentricityResonancOperator):
    def __init__(self,initial_state,dt,j,indexIn=1,indexOut=2,Lambda0=None):
        res_vec = np.array([2-j,j]) / 2
        pIn = initial_state.particles[indexIn]
        pOut = initial_state.particles[indexOut]
        self.mIn = pIn.m
        self.MIn = pIn.M
        self.mOut = pOut.m
        self.MOut = pOut.M
        self.j = j
        if Lambda0 is None:
            _pvars = Poincare(initial_state.G,[pIn,pOut])
            _,Lambda0 = get_res_chain_reference_Lambdas_and_semimajor_axes(_pvars,[(j-2)/2])
            _,self.Lambda0In,self.Lambda0Out = Lambda0
        else:
            self.Lambda0In = Lambda0[indexIn]
            self.Lambda0Out = Lambda0[indexOut]
        Amtrx = get_second_order_eccentricity_resonance_matrix(
                self.j,
                initial_state.G,
                self.mIn,
                self.mOut,
                self.MIn,
                self.MOut,
                self.Lambda0In,
                self.Lambda0Out
        )
        bvec = np.zeros(2)
        super(SecondOrderEccentricityResonanceOperator,self).__init__(initial_state,indexIn,indexOut,res_vec,Amtrx,bvec,dt)

class SecondOrderInclinationResonanceOperator(LinearInclinationResonancOperator):
    def __init__(self,initial_state,dt,j,indexIn=1,indexOut=2,Lambda0=None):
        res_vec = np.array([2-j,j]) / 2
        pIn = initial_state.particles[indexIn]
        pOut = initial_state.particles[indexOut]
        self.mIn = pIn.m
        self.MIn = pIn.M
        self.mOut = pOut.m
        self.MOut = pOut.M
        self.j = j
        if Lambda0 is None:
            _pvars = Poincare(initial_state.G,[pIn,pOut])
            _,Lambda0 = get_res_chain_reference_Lambdas_and_semimajor_axes(_pvars,[(j-2)/2])
            _,self.Lambda0In,self.Lambda0Out = Lambda0
        else:
            self.Lambda0In = Lambda0[indexIn]
            self.Lambda0Out = Lambda0[indexOut]
        Amtrx = get_second_order_inclination_resonance_matrix(
                self.j,
                initial_state.G,
                self.mIn,
                self.mOut,
                self.MIn,
                self.MOut,
                self.Lambda0In,
                self.Lambda0Out
        )
        bvec = np.zeros(2)
        super(SecondOrderInclinationResonanceOperator,self).__init__(initial_state,indexIn,indexOut,res_vec,Amtrx,dt)


from .poisson_series import DFTermSeries
from scipy.optimize import root
from .disturbing_function import ResonanceTermsList, SecularTermsList

class MeanMotionResonanceDFTermsEvolutionOperator(EvolutionOperator):
    """
    Evolution operator for a collection of disturbing function
    terms associated with a specific mean motion resonance between
    a pair of planets. 
    
    The evolution of inclination/eccentricity variables are
    computed by the implicit midpoint method. This is a second order
    symplectic method. The evolution of semi-major axes are computed
    so that both the planet pair's angular momentum and the conserved
    quantity associated with the mean motion resonance are conserved.
    The planets' mean longtiudes do not evolve.

    Arguments
    ---------
    initial state : celmech.Poincare
        Poincare variables state from which the operator is initialized.
    dt : float
        Time-step of the operator.
    j : int
        Integer determining the resonance as the j:j-k resonance.
    k : int
        The order of the resonance.
    terms_list : list
        List of distrubing function terms to include.
        List items are tuples of the form (kvec, zvec).
    indexIn : int
        Index of the inner planet particle in the resonance.
    indexOut : int
        Index of the outer planet particle in the resonance.
    Lambda0 : array-like
        The Poincare momenta Lambda are treated as constant
        in the disturing function. Lambda0 sets the values
        of these constant momenta and should be an array
        with an entry for each particle in the system.
        If no value is supplied, initial values are chosen
        at the center of exact resonance using conservation of
        angular momentum and (j-k) * LambdaOut + j * LambdaIn
        presuming only two planets are in the system.
    """

    def __init__(self, initial_state, dt, j, k, terms_list, indexIn=1, indexOut=2, Lambda0=None,**kwargs):

        pIn = initial_state.particles[indexIn]
        pOut = initial_state.particles[indexOut]
        self.indexIn = indexIn
        self.indexOut = indexOut
        self.mIn = pIn.m
        self.MIn = pIn.M
        self.mOut = pOut.m
        self.MOut = pOut.M
        self.G = initial_state.G
        
        if Lambda0 is None:
            _pvars = Poincare(initial_state.G,[pIn,pOut])
            _,Lambda0 = get_res_chain_reference_Lambdas_and_semimajor_axes(_pvars,[(j-k)/k])
            _,self.Lambda0In,self.Lambda0Out = Lambda0
        else:
            self.Lambda0In = Lambda0[indexIn]
            self.Lambda0Out = Lambda0[indexOut]

        self.rtLambda0_inv = 1 / sqrt([self.Lambda0In,self.Lambda0Out])
        self.qp_to_XY_factors = np.concatenate((self.rtLambda0_inv,0.5*self.rtLambda0_inv))
        self.DF_prefactor = -self.G**2*self.MOut**2*self.mOut**3 * ( self.mIn / self.MIn) / (self.Lambda0Out**2)
        self._dt = dt
        self._h = self.DF_prefactor * self._dt
        self.alpha = ((j-k)/j)**(2/3)
        self.terms_list = terms_list 
        self.DFSeries = DFTermSeries.from_resonance_list(
                self.terms_list,
                self.alpha,
                [self.Lambda0In,self.Lambda0Out]
        )
        self.res_vec = np.array([k - j , j])
        self.Lambdas_vec = np.array([-1 * self.res_vec[1] , self.res_vec[0]])
        self.Lambdas_Mtrx = np.linalg.inv([ self.Lambdas_vec , [1,1]])
    
    @classmethod
    def from_Order_Range(cls, initial_state, dt, j, k, Nmin, Nmax, indexIn=1, indexOut=2, Lambda0=None, include_secular_terms = False):
        terms = ResonanceTermsList(j,k,Nmin,Nmax)
        if include_secular_terms:
            terms += SecularTermsList(Nmin,Nmax)
        return cls(initial_state, dt, j, k, terms, indexIn, indexOut, Lambda0)

    @property
    def dt(self):
        return super().dt

    @dt.setter
    def dt(self,value):
        self._dt = value
        self._h = self.DF_prefactor * self._dt 

    @property
    def indices(self):
        return self.indexIn-1,self.indexOut-1

    def state_vec_to_qp_vec_and_lambdaLambda(self,state_vec):
        vecs = self._state_vector_to_individual_vectors(state_vec)
        kappa = vecs[self.indices,0] 
        eta = vecs[self.indices,1] 
        sigma = vecs[self.indices,4] 
        rho = vecs[self.indices,5] 
        lambdas = vecs[self.indices,3]
        Lambdas = vecs[self.indices,2]
        
        return np.concatenate((eta,rho,kappa,sigma)),lambdas,Lambdas

    def deriv_and_jacobian_from_qp_vec(self,qp_vec,lambda_arr):  
        XYvec = (qp_vec[4:] - 1j * qp_vec[:4]) * self.qp_to_XY_factors
        _,qp_dot,qp_dotJac,_ = self.DFSeries._evaluate_with_jacobian(lambda_arr,XYvec)
        return qp_dot, qp_dotJac

    def implicit_midpoint_f_and_Df(self,qp_vec1,qp_vec0,lambda_arr):
        h = self._h
        qp_vec_mid = 0.5 * (qp_vec1+qp_vec0)
        qp_dot,qp_dotJac = self.deriv_and_jacobian_from_qp_vec(qp_vec_mid,lambda_arr)
        f = qp_vec1 - qp_vec0 - h * qp_dot
        Df = np.eye(8) - 0.5 * h * qp_dotJac
        return f,Df

    def implicit_midpoint_step(self,qp_vec,lambda_arr):
        rt = root(
                self.implicit_midpoint_f_and_Df,
                qp_vec,
                args = (qp_vec,lambda_arr),
                jac=True
            )
        return rt.x

    def apply(self):
        pass

    def apply_to_state_vector(self,state_vec):
        vecs = self._state_vector_to_individual_vectors(state_vec)
        kappa = vecs[self.indices,0] 
        eta = vecs[self.indices,1] 
        sigma = vecs[self.indices,4] 
        rho = vecs[self.indices,5] 
        lambdas = vecs[self.indices,3]
        Lambdas = vecs[self.indices,2]
        qp_vec =  np.concatenate((eta,rho,kappa,sigma))
        AMD0 = 0.5 * np.sum(qp_vec**2)
        C1 = self.Lambdas_vec @ Lambdas
        C2 = np.sum(Lambdas) - AMD0
        qp_vec_new = self.implicit_midpoint_step(qp_vec,lambdas)
        AMD1 = 0.5 * np.sum(qp_vec_new**2)
        Lambdas_new = self.Lambdas_Mtrx @ np.array([C1, C2 + np.sum(AMD1)])
        
        vecs[self.indices,2] = Lambdas_new

        # eta
        vecs[self.indices,1] = qp_vec_new[0:2]  

        # kappa
        vecs[self.indices,0] = qp_vec_new[4:6]  
        
        # rho
        vecs[self.indices,5] = qp_vec_new[2:4]  
        
        # sigma
        vecs[self.indices,4] = qp_vec_new[6:8]  
        
        return vecs.reshape(-1)

    
def get_res_chain_reference_Lambdas_and_semimajor_axes(pvars,slist):
    coeffs = np.zeros(pvars.N)
    alpha_inv = np.zeros(pvars.N)
    alpha_inv[1] = 1    
    coeffs[1] = 1 + slist[0]
    ps = pvars.particles
    mu = np.array([p.mu for p in ps])
    M = np.array([p.M for p in ps])
    tot = coeffs[1] * mu[1] * sqrt(M[1])
    for i in range(2,len(coeffs)):
        coeffs[i] = coeffs[i-1] * slist[i-2] / (1 + slist[i-2])
        alpha_inv[i] = alpha_inv[i-1]  * ((1 + slist[i-2]) / slist[i-2])**(2/3)
        tot += coeffs[i] * mu[i] * sqrt(M[i] * alpha_inv[i])        
    Lvals = np.append([0],[p.Lambda for p in ps[1:]])
    C = coeffs @ Lvals
    a10 = (C / tot)**2
    a0 = a10 * alpha_inv 
    L0 = mu * sqrt(M * a0)
    return a0,L0

def get_first_order_eccentricity_resonance_matrix_and_vector(j,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out):
    r"""
    Get matrix A and vector b that define the Hamiltonian terms

    .. math::

        H_{res} = \frac{e^{2i\theta}}{2} \left( \bar{x} \cdot A \cdot \bar{x} + b e^{-i\theta} \cdot \bar{x} \right)

    associated with a first-order MMR in the disturbing function expansion.

    Arguments
    ---------
    j : int
        Determines the j:j-1 first order MMR.
    G : float
        Gravitational constant.
    mIn : float
        Mass of inner planet
    mOut : float
        Mass of outer planet
    MIn : float
        Total mass of star + inner planet
    Mout : float
        Total mass of star + outer planet
    Lambda0In : float
        Reference value of Lambda for inner planet
    Lambda0Out : float
        Reference value of Lambda for outer planet
    Returns
    -------
    A : sympy.Matrix , shape (2,2)
        The matrix A
    b : sympy.Matrix, shape (2,)
        The vector b

    """

    zs = [0,0,0,0]
    muIn = mIn * (MIn - mIn) / MIn
    muOut = mOut * (MOut - mOut) / MOut
    aIn0 = (Lambda0In / muIn)**2 / MIn / G
    aOut0 = (Lambda0Out / muOut)**2 / MOut / G
    alpha0 = aIn0 / aOut0
    aOut_inv = G*MOut*muOut*muOut / LambdaOut / LambdaOut  
    prefactor = -G * mIn * mOut * aOut_inv
    A = get_second_order_eccentricity_resonance_matrix(2 * j,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out)
    b = np.zeros(2)
    js = [j,1-j,-1,0,0,0]
    b[0]  = eval_DFCoeff_dict(DFCoeff_C(*js,*zs),alpha0)
    js = [j,1-j,0,-1,0,0]
    b[1] = eval_DFCoeff_dict(DFCoeff_C(*js,*zs),alpha0)
    # scale X --> x
    scaleMtrx = np.diag([sqrt(2/Lambda0In),sqrt(2/Lambda0Out)])
    b = scaleMtrx @ b
    return  A, prefactor * b

def get_second_order_eccentricity_resonance_matrix(j,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out):
    zs = [0,0,0,0]
    muIn = mIn * (MIn - mIn) / MIn
    muOut = mOut * (MOut - mOut) / MOut
    aIn0 = (Lambda0In / muIn)**2 / MIn / G
    aOut0 = (Lambda0Out / muOut)**2 / MOut / G
    alpha0 = aIn0 / aOut0
    aOut_inv = G*MOut*muOut*muOut / LambdaOut / LambdaOut  
    prefactor = -G * mIn * mOut * aOut_inv
    params = dict()
    A = np.zeros((2,2))

    js = [j,2-j,-2,0,0,0]
    A[0,0]  = eval_DFCoeff_dict(DFCoeff_C(*js,*zs),alpha0) 

    js = [j,2-j,0,-2,0,0]
    A[1,1]  = eval_DFCoeff_dict(DFCoeff_C(*js,*zs),alpha0)

    js = [j,2-j,-1,-1,0,0]
    A[1,0] = eval_DFCoeff_dict(DFCoeff_C(*js,*zs),alpha0) / 2
    A[0,1] = A[1,0]
    # scale X --> x
    scaleMtrx = np.diag([sqrt(2/Lambda0In),sqrt(2/Lambda0Out)])
    A = scaleMtrx @ A @ scaleMtrx
    return prefactor * A

def get_second_order_inclination_resonance_matrix(j,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out):
    zs = [0,0,0,0]
    muIn = mIn * (MIn - mIn) / MIn
    muOut = mOut * (MOut - mOut) / MOut
    aIn0 = (Lambda0In / muIn)**2 / MIn / G
    aOut0 = (Lambda0Out / muOut)**2 / MOut / G
    alpha0 = aIn0 / aOut0
    aOut_inv = G*MOut*muOut*muOut / LambdaOut / LambdaOut  
    prefactor = -G * mIn * mOut * aOut_inv
    params = dict()
    A = np.zeros((2,2))

    js = [j,2-j,0,0,-2,0]
    A[0,0]  = eval_DFCoeff_dict(DFCoeff_C(*js,*zs),alpha0) 

    js = [j,2-j,0,0,0,-2]
    A[1,1]  = eval_DFCoeff_dict(DFCoeff_C(*js,*zs),alpha0)

    js = [j,2-j,0,0,-1,-1]
    A[1,0] = eval_DFCoeff_dict(DFCoeff_C(*js,*zs),alpha0) / 2
    A[0,1] = A[1,0]
    # scale X --> x
    scaleMtrx = np.diag([sqrt(0.5 / Lambda0In),sqrt(0.5 / Lambda0Out)])
    A = scaleMtrx @ A @ scaleMtrx
    
    return prefactor * A




import rebound as rb
import reboundx
import numpy as np
import theano
import theano.tensor as T
from warnings import warn
from scipy.optimize import lsq_linear
from scipy.integrate import odeint
import warnings
from tqdm import tqdm

from .numerical_resonance_utils import planar_els2xv,calc_Hint_components_spatial,calc_Hint_components_planar
from .nbody_simulation_utilities import get_canonical_heliocentric_orbits, add_canonical_heliocentric_elements_particle, align_simulation
from .miscellaneous import getOmegaMatrix
# set to fast compile, for testing at least...
######################
####### PLANAR #######
######################
def _get_planar_compiled_Hpert_full():
        # Planet masses: m1,m2
        m1,m2 = T.dscalars(2)
        mstar = 1
        mu1  = m1 * mstar / (mstar  + m1) 
        mu2  = m2 * mstar / (mstar  + m2) 
        Mstar1 = mstar + m1
        Mstar2 = mstar + m2
        beta1 = mu1 * T.sqrt(Mstar1/mstar) / (mu1 + mu2)
        beta2 = mu2 * T.sqrt(Mstar2/mstar) / (mu1 + mu2)
        j,k = T.lscalars('jk')
        s = (j-k) / k

        # Dynamical variables:
        Ndof = 2
        Nconst = 1
        psi = T.scalar()
        dyvars = T.vector()
        y1, y2, x1, x2, amd = [dyvars[i] for i in range(2*Ndof + Nconst)]

        a20 = T.constant(1.)
        a10 = ((j-k)/j)**(2/3) * (Mstar1 / Mstar2)**(1/3)
        L10 = beta1 * T.sqrt(a10)
        L20 = beta2 * T.sqrt(a20)
        Ltot = L10 + L20
        f = L10/L20
        L2res = (Ltot + amd) / (1+f)
        Psi = -k * (s * L2res + (1+s) * f * L2res) 
        
        ###
        # actions
        ###
        I1 = 0.5 * (x1*x1 + y1*y1)
        I2 = 0.5 * (x2*x2 + y2*y2)
        L1 = -s*Ltot - Psi/k - s * (I1 + I2)
        L2 = (1+s)*Ltot + Psi/k + (1+s) * (I1 + I2)

        # Set lambda2=0
        l2 = T.constant(0.)
        l1 = -1 * k * psi 
        theta_res = (1+s) * l2 - s * l1
        cos_theta_res = T.cos(theta_res)
        sin_theta_res = T.sin(theta_res)
        
        kappa1 = x1 * cos_theta_res + y1 * sin_theta_res
        eta1   = y1 * cos_theta_res - x1 * sin_theta_res
        
        kappa2 = x2 * cos_theta_res + y2 * sin_theta_res
        eta2   = y2 * cos_theta_res - x2 * sin_theta_res
        
        
        Xre1 = kappa1 / T.sqrt(L1)
        Xim1 = -eta1 / T.sqrt(L1)

        Xre2 = kappa2 / T.sqrt(L2)
        Xim2 = -eta2 / T.sqrt(L2)

        absX1_sq = 2 * I1 / L1
        absX2_sq = 2 * I2 / L2
        X_to_z1 = T.sqrt(1 - absX1_sq / 4 )
        X_to_z2 = T.sqrt(1 - absX2_sq / 4 )

        a1 = (L1 / beta1 )**2 
        k1 = Xre1 * X_to_z1
        h1 = Xim1 * X_to_z1
        e1 = T.sqrt( absX1_sq ) * X_to_z1

        a2 = (L2 / beta2 )**2 
        k2 = Xre2 * X_to_z2
        h2 = Xim2 * X_to_z2

        Hdir,Hind = calc_Hint_components_planar(
                a1,a2,l1,l2,h1,k1,h2,k2,Mstar1/mstar,Mstar2/mstar
        )
        eps = m1*m2/ (mu1 + mu2) / T.sqrt(mstar)
        extra_ins = [m1,m2,j,k]
        ins = [psi] + [dyvars] + extra_ins
        ################################
        Hpert = eps * (Hdir + Hind/mstar)
        gradHpert = T.grad(Hpert,wrt=dyvars)
        ################################
        # Scalars
        Hpert_fn = theano.function(
            inputs=ins,
            outputs=Hpert,
            givens=None,
            on_unused_input='ignore'
        )
        # Gradients
        gradHpert_fn = theano.function(
            inputs=ins,
            outputs=gradHpert,
            givens=None,
            on_unused_input='ignore'
        )
        return Hpert_fn, gradHpert_fn

def _get_planar_compiled_theano_functions(N_QUAD_PTS):
        # Planet masses: m1,m2
        m1,m2 = T.dscalars(2)
        mstar = 1
        mu1  = m1 * mstar / (mstar  + m1) 
        mu2  = m2 * mstar / (mstar  + m2) 
        Mstar1 = mstar + m1
        Mstar2 = mstar + m2
        beta1 = mu1 * T.sqrt(Mstar1/mstar) / (mu1 + mu2)
        beta2 = mu2 * T.sqrt(Mstar2/mstar) / (mu1 + mu2)
        j,k = T.lscalars('jk')
        s = (j-k) / k
        
        # Angle variable for averaging over
        psi = T.dvector()

        # Quadrature weights
        quad_weights = T.dvector('w')

        # Dynamical variables:
        Ndof = 2
        Nconst = 1
        dyvars = T.vector()
        y1, y2, x1, x2, amd = [dyvars[i] for i in range(2*Ndof + Nconst)]

        a20 = T.constant(1.)
        a10 = ((j-k)/j)**(2/3) * (Mstar1 / Mstar2)**(1/3)
        L10 = beta1 * T.sqrt(a10)
        L20 = beta2 * T.sqrt(a20)
        Ltot = L10 + L20
        f = L10/L20
        L2res = (Ltot + amd) / (1+f)
        Psi = -k * (s * L2res + (1+s) * f * L2res) 
        
        ###
        # actions
        ###
        I1 = 0.5 * (x1*x1 + y1*y1)
        I2 = 0.5 * (x2*x2 + y2*y2)
        L1 = -s*Ltot - Psi/k - s * (I1 + I2)
        L2 = (1+s)*Ltot + Psi/k + (1+s) * (I1 + I2)

        # Set lambda2=0
        l2 = T.constant(0.)
        l1 = -1 * k * psi 
        theta_res = (1+s) * l2 - s * l1
        cos_theta_res = T.cos(theta_res)
        sin_theta_res = T.sin(theta_res)
        
        kappa1 = x1 * cos_theta_res + y1 * sin_theta_res
        eta1   = y1 * cos_theta_res - x1 * sin_theta_res
        
        kappa2 = x2 * cos_theta_res + y2 * sin_theta_res
        eta2   = y2 * cos_theta_res - x2 * sin_theta_res
        
        
        Xre1 = kappa1 / T.sqrt(L1)
        Xim1 = -eta1 / T.sqrt(L1)

        Xre2 = kappa2 / T.sqrt(L2)
        Xim2 = -eta2 / T.sqrt(L2)

        absX1_sq = 2 * I1 / L1
        absX2_sq = 2 * I2 / L2
        X_to_z1 = T.sqrt(1 - absX1_sq / 4 )
        X_to_z2 = T.sqrt(1 - absX2_sq / 4 )

        a1 = (L1 / beta1 )**2 
        k1 = Xre1 * X_to_z1
        h1 = Xim1 * X_to_z1
        e1 = T.sqrt( absX1_sq ) * X_to_z1

        a2 = (L2 / beta2 )**2 
        k2 = Xre2 * X_to_z2
        h2 = Xim2 * X_to_z2
        e2 = T.sqrt( absX2_sq ) * X_to_z2

        beta1p = T.sqrt(Mstar1) * beta1
        beta2p = T.sqrt(Mstar2) * beta2
        Hkep = -0.5 * beta1p / a1 - 0.5 * beta2p / a2
        
        Hdir,Hind = calc_Hint_components_planar(
                a1,a2,l1,l2,h1,k1,h2,k2,Mstar1,Mstar2
        )
        eps = m1*m2/ (mu1 + mu2) / T.sqrt(mstar)
        Hpert = eps * (Hdir + Hind/mstar)
        Hpert_av = Hpert.dot(quad_weights)
        Htot = Hkep + Hpert_av

        ######################
        # Dissipative dynamics
        ######################
        tau_alpha_0, K1, K2, p = T.dscalars(4)
        y1dot_dis,y2dot_dis,x1dot_dis,x2dot_dis,amddot_dis = T.dscalars(5)
        tau_m_inv = 1/tau_alpha_0
        alpha_res = a10/a20
        # Define timescales
        tau_e1 = tau_alpha_0 / K1
        tau_e2 = tau_alpha_0 / K2
        tau_a1_0_inv = -beta2p * alpha_res * tau_m_inv / (beta1p + alpha_res * beta2p)
        tau_a2_0_inv = beta1p * tau_m_inv / (beta1p + alpha_res * beta2p)
        tau_a1 = 1 / (tau_a1_0_inv + 2 * p * e1*e1 / tau_e1 )
        tau_a2 = 1 / (tau_a2_0_inv + 2 * p * e2*e2 / tau_e2 )
        
        tau_L1 = 2 * tau_a1
        tau_L2 = 2 * tau_a2
        Gamma1=I1
        Gamma2=I2
        tau_Gamma1_inv = 1/tau_L1 + (Gamma1-2*L1) / (Gamma1-L1) / tau_e1 
        tau_Gamma2_inv = 1/tau_L2 + (Gamma2-2*L2) / (Gamma2-L2) / tau_e2 
        # Time derivatives of canonical variables
        x1dot_dis = -0.5 * x1 * tau_Gamma1_inv
        x2dot_dis = -0.5 * x2 * tau_Gamma2_inv
        y1dot_dis = -0.5 * y1 * tau_Gamma1_inv
        y2dot_dis = -0.5 * y2 * tau_Gamma2_inv

        L1dot_dis = -1 * L1/tau_L1
        L2dot_dis = -1 * L2/tau_L2
        Gamma1dot_dis = -1 * I1 * tau_Gamma1_inv 
        Gamma2dot_dis = -1 * I2 * tau_Gamma2_inv 
        Ltot_dot_dis = L1dot_dis + L2dot_dis - Gamma1dot_dis - Gamma2dot_dis
        Psi_dot_dis = -k * (s * L2dot_dis + (1+s) * L1dot_dis) 
        amddot_dis = -1 * Ltot_dot_dis - Psi_dot_dis * (1+f) /(s + f*(1+s)) / k
        
        #####################################################
        # Set parameters for compiling functions with Theano
        #####################################################
        
        # Get numerical quadrature nodes and weight
        nodes,weights = np.polynomial.legendre.leggauss(N_QUAD_PTS)
        
        # Rescale for integration interval from [-1,1] to [-pi,pi]
        nodes = nodes * np.pi
        weights = weights * 0.5
        
        # 'givens' will fix some parameters of Theano functions compiled below
        givens = [(psi,nodes),(quad_weights,weights)]

        # 'ins' will set the inputs of Theano functions compiled below
        #   Note: 'extra_ins' will be passed as values of object attributes
        #   of the 'ResonanceEquations' class defined below
        extra_ins = [m1,m2,j,k,tau_alpha_0,K1,K2,p]
        ins = [dyvars] + extra_ins
        
        # Define flows and jacobians.

        #  Conservative flow
        gradHtot = T.grad(Htot,wrt=dyvars)
        gradHpert = T.grad(Hpert_av,wrt=dyvars)
        gradHkep = T.grad(Hkep,wrt=dyvars)

        hessHtot = theano.gradient.hessian(Htot,wrt=dyvars)
        hessHpert = theano.gradient.hessian(Hpert_av,wrt=dyvars)
        hessHkep = theano.gradient.hessian(Hkep,wrt=dyvars)

        Jtens = T.as_tensor(np.pad(getOmegaMatrix(Ndof),(0,Nconst),'constant'))
        H_flow_vec = Jtens.dot(gradHtot)
        Hpert_flow_vec = Jtens.dot(gradHpert)
        Hkep_flow_vec = Jtens.dot(gradHkep)

        H_flow_jac = Jtens.dot(hessHtot)
        Hpert_flow_jac = Jtens.dot(hessHpert)
        Hkep_flow_jac = Jtens.dot(hessHkep)
        
        # Dissipative flow
        dis_flow_vec = T.stack(y1dot_dis,y2dot_dis,x1dot_dis,x2dot_dis,amddot_dis)
        dis_flow_jac = theano.gradient.jacobian(dis_flow_vec,dyvars)

        # Extras
        sigma1 = T.arctan2(y1,x1)
        sigma2 = T.arctan2(y2,x2)
        orbels = [a1,e1,k*sigma1,a2,e2,k*sigma2]
        dis_timescales = [1/tau_a1_0_inv,1/tau_a2_0_inv,tau_e1,tau_e2]

        orbels_dict = dict(zip(
                ['a1','e1','theta1','a2','e2','theta2'],
                orbels
            )
        )

        actions = [L1,L2,I1,I2]
        actions_dict = dict(
            zip(
                ['L1','L2','Gamma1','Gamma2'],
                actions
                )
        )
        
        timescales_dict = dict(zip(
            ['tau_m1','tau_m2','tau_e1','tau_e2'],
            dis_timescales
            )
        )
        ##########################
        # Compile Theano functions
        ##########################
        func_dict={
         # Hamiltonians
         'H':Htot,
         'Hpert':Hpert_av,
         'Hkep':Hkep,
         # Hamiltonian flows
         'H_flow':H_flow_vec,
         'Hpert_flow':Hpert_flow_vec,
         'Hkep_flow':Hkep_flow_vec,
         # Hamiltonian flow Jacobians
         'H_flow_jac':H_flow_jac,
         'Hpert_flow_jac':Hpert_flow_jac,
         'Hkep_flow_jac':Hkep_flow_jac,
         # Dissipative flow and Jacobian
         'dissipative_flow':dis_flow_vec,
         'dissipative_flow_jac':dis_flow_jac,
         # Extras
         'orbital_elements':orbels_dict,
         'actions':actions_dict,
         'timescales':timescales_dict
        }
        compiled_func_dict=dict()
        for key,val in func_dict.items():
            if key == 'timescales':
                inputs = extra_ins
            else:
                inputs = ins 
            cf = theano.function(
                inputs=inputs,
                outputs=val,
                givens=givens,
                on_unused_input='ignore'
            )
            compiled_func_dict[key]=cf
        return compiled_func_dict
        

class PlanarResonanceEquations():
    r"""
    A class for the model describing the dynamics of a pair of planar planets
    in/near a mean motion resonance.

    Includes the effects of dissipation.

    Attributes
    ----------
    j : int
        Together with k specifies j:j-k resonance
    
    k : int
        Order of resonance.
    
    alpha : float
        Semi-major axis ratio :math:`a_1/a_2`

    eps : float
        Mass parameter m1*mu2 / (mu1+mu2)

    m1 : float
        Inner planet mass

    m2 : float
        Outer planet mass

    tau_alpha : float
        Migration timescale defined by :math:`\tau_\alpha^{-1} =\tau_{m,2}^{-1} - \tau_{m,1}^{-1}`

    K1 : float
        Ratio of inner planet eccentricity damping timescale to migration timescale, tau_alpha.

    K2 : float
        Ratio of outer planet eccentricity damping timescale to migration timescale, tau_alpha.

    p : float
        Sets coupling between eccentricity damping and semi-major axis damping so that

        .. math::

            \frac{d}{dt}a_i = -a_i/\tau_{m,i} - 2pe_i^2/\tau_{e,i} 
        
        A value of p=1 corresponds to eccentricity damping at constant angular momentum.

    timescales : dict
        Dictionary containing the migration and eccentricity damping time-scales of 
        the inner and outer planets based on the damping parameters of the resonance
        model.

    """
    def __init__(self,j,k, n_quad_pts = 40, m1 = 1e-5 , m2 = 1e-5,K1=100, K2=100, tau_alpha = 1e5, p = 1):
        self.j = j
        self.k = k
        self.m1 = m1
        self.m2 = m2
        self.K1 = K1
        self.K2 = K2
        self.tau_alpha = tau_alpha
        self.p = p 
        self.n_quad_pts = n_quad_pts
        self._Hpert_full,self._grad_Hpert_full = _get_planar_compiled_Hpert_full()
        self._funcs = _get_planar_compiled_theano_functions(n_quad_pts)

    @property
    def extra_args(self):
        return [self.m1,self.m2,self.j,self.k,self.tau_alpha,self.K1,self.K2,self.p]

    @property
    def extra_args_short(self):
        return [self.m1,self.m2,self.j,self.k]
    @property
    def mu1(self):
        return self.m1 / (1 + self.m1)
    @property
    def mu2(self):
        return self.m2 / (1 + self.m2)
    @property
    def beta1(self):
        return self.mu1 * np.sqrt(1+self.m1) / (self.mu1 + self.mu2)
    @property
    def beta2(self):
        return self.mu2 * np.sqrt(1+self.m2) / (self.mu1 + self.mu2)
    @property
    def eps(self):
        return self.m1 * self.m2 / (self.mu1 + self.mu2)
    @property
    def alpha(self):
        alpha0 = ((self.j-self.k)/self.j)**(2/3)
        return alpha0 * ((1 + self.m1) / (1+self.m2))**(1/3)
    @property
    def timescales(self):
        return self._funcs['timescales'](*self.extra_args)

    
    def H(self,z):
        """
        Calculate the value of the Hamiltonian.

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            The value of the Hamiltonian evaluated at z.
        """
        return self._funcs['H'](z,*self.extra_args)

    def H_kep(self,z):
        """
        Calculate the value of the Keplerian component of the Hamiltonian.

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            The value of the Keplerian part of the Hamiltonian evaluated at z.
        """
        return self._funcs['Hkep'](z,*self.extra_args)

    def H_kep_flow(self,z):
        """
        Calculate the flow vector generated by the Keplerian component of the Hamiltonian.

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        ndarray : 
            The flow vector of the Keplerian component of the Hamiltonian
        """
        return self._funcs['Hkep_flow'](z,*self.extra_args)

    def H_pert(self,z):
        r"""
        Calculate the value of the perturbation component of the Hamiltonian.

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            The value of the perturbation part of the Hamiltonian evaluated at z.
        """
        return self._funcs['Hpert'](z,*self.extra_args)

    def H_pert_flow(self,z):
        r"""
        Calculate the flow vector generated by the perturbation component of the Hamiltonian.

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        ndarray : 
            The flow vector generated by the perturbation part of the Hamiltonian evaluated at z.
        """
        return self._funcs['Hpert_flow'](z,*self.extra_args)

    def H_pert_full(self,psi,z):
        r"""
        Calculate the value of the perturbation component of the full, unaveraged
        Hamiltonian.

        Arguments
        ---------
        psi : float
            Angle variable 

            .. math::
                (\lambda_2-\lambda_1)/k  

        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            The value of the perturbation part of the Hamiltonian evaluated at :math:`(\psi,z)`.
        """
        return self._Hpert_full(psi,z,*self.extra_args_short)

    def H_pert_osc(self,psi,z):
        r"""
        Calculate the oscillating part of value of the perturbation component of the unaveraged Hamiltonian.

        Arguments
        ---------
        psi : float
            Angle variable 

            .. math::
                (\lambda_2-\lambda_1)/k  

        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            The value of the perturbation part of the Hamiltonian evaluated at :math:`(\psi,z)`.
        """
        return self.H_pert_full(psi,z) - self.H_pert(z)

    def grad_H_pert_full(self,psi,z):
        r"""
        Calculate the gradient of the perturbation component of the full, unaveraged
        Hamiltonian with respect to variables :math:`z`.

        Arguments
        ---------
        psi : float
            Angle variable 

            .. math::
                (\lambda_2-\lambda_1)/k  

        z : ndarray
            Dynamical variables

        Returns
        -------
        grad : ndarray
            The gradient the perturbation part of the Hamiltonian with respect to 
            :math:`z` evaluated at :math:`(\psi,z)`.
        """
        return self._grad_Hpert_full(psi,z,*self.extra_args_short)

    def H_flow(self,z):
        r"""
        Calculate flow induced by the Hamiltonian

        .. math::
            \dot{z} = \Omega \cdot \nabla_{z}H(z)

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        ndarray : 
            Flow vector
        """
        return self._funcs['H_flow'](z,*self.extra_args)


    def H_flow_jac(self,z):
        r"""
        Calculate the Jacobian of the flow induced by the Hamiltonian

        .. math::
             \Omega \cdot \Delta H(z)

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        ndarray : 
            Jacobian matrix
        """
        return self._funcs['H_flow_jac'](z,*self.extra_args)

    def flow(self,z):
        r"""
        Calculate the flow vector of the equations
        of motion

        .. math::
            \dot{z} = \Omega \cdot \nabla_{z}H(z) + f_{dis}(z)

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        ndarray : 
            Flow vector
        """
        return self._funcs['H_flow'](z,*self.extra_args) + self._funcs['dissipative_flow'](z,*self.extra_args)

    def flow_jac(self,z):
        r"""
        Calculate the Jacobian of the equations of motion
        given by 

        .. math::
             \Omega \cdot \Delta H(z) + \nabla f_{dis}(z)

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        ndarray : 
            Jacobian matrix
        """
        return self._funcs['H_flow_jac'](z,*self.extra_args) + self._funcs['dissipative_flow_jac'](z,*self.extra_args)

    def omega_syn(self,z):
        r"""
        Calculate the synodic frequency,

        .. math::
            \omega_{syn} = n_2 - n_1

        as a function of dynamical variables :math:`z`

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            Synodic frequency
        """
        actions = self.dyvars_to_Poincare_actions(z)
        L1,L2 = actions['L1'],actions['L2']
        n1 = np.sqrt(1 + self.m1) * self.beta1**3 / L1**3
        n2 = np.sqrt(1 + self.m2) * self.beta2**3 / L2**3
        return n2 - n1

    def grad_omega_syn(self,z):
        r"""
        Calculate the gradient of the synodic frequency,

        .. math::
            \nabla_{\pmb{z}} \omega_{syn} 

        as a function of dynamical variables :math:`\pmb{z}`

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        float  
            Synodic frequency
        """
        s  = (self.j - self.k) / self.k
        y1,y2,x1,x2,amd = z
        actions = self.dyvars_to_Poincare_actions(z)
        L1,L2 = actions['L1'],actions['L2']
        n1 = np.sqrt(1 + self.m1) * self.beta1**3 / L1**3
        n2 = np.sqrt(1 + self.m2) * self.beta2**3 / L2**3
        dn1_dL1 = -3 * n1 / L1
        dn2_dL2 = -3 * n2 / L2

        dL1dx1 = -1 * s * x1
        dL1dy1 = -1 * s * y1
        dL1dx2 = -1 * s * x2
        dL1dy2 = -1 * s * y2

        dL2dx1 = (1+s) *  x1
        dL2dy1 = (1+s) *  y1
        dL2dx2 = (1+s) *  x2
        dL2dy2 = (1+s) *  y2
        
        return np.array((
            dn2_dL2 *  dL2dy1 - dn1_dL1 * dL1dy1,
            dn2_dL2 *  dL2dy2 - dn1_dL1 * dL1dy2,
            dn2_dL2 *  dL2dx1 - dn1_dL1 * dL1dx1,
            dn2_dL2 *  dL2dx2 - dn1_dL1 * dL1dx2
            ))

    def dyvars_to_orbital_elements(self,z):
        r"""
        Convert dynamical variables

        .. math::
            z = (y_1,y_2,x_1,x_2,{\cal D})

        to orbital elements

        .. math::
            (a_1,e_1,\theta_1,a_2,e_2,\theta_2)
        """

        return self._funcs['orbital_elements'](z,*self.extra_args)

    def orbital_elements_to_dyvars(self,orbels):
        r"""
        Convert orbital elements

        .. math::
            (a_1,e_1,\theta_1,a_2,e_2,\theta_2)

        to dynamical variables

        .. math::
            z = (y_1,y_2,x_1,x_2,{\cal D})
        """
        # Total angular momentum constrained by
        # Ltot = beta1 * sqrt(alpha_res) + beta2 - amd
        Mstar = 1
        m1 = self.m1
        m2 = self.m2
        k = self.k
        j = self.j
        alpha_res = self.alpha
    
        s = (j - k) / k
        
        keys=['a1','e1','theta1','a2','e2','theta2']
        a1,e1,theta1,a2,e2,theta2 = [orbels[key] for key in keys]
        
        L1 = self.beta1 * np.sqrt(a1)
        L2 = self.beta2 * np.sqrt(a2)
        Gamma1 = L1 * (1 - np.sqrt(1-e1**2))
        Gamma2 = L2 * (1 - np.sqrt(1-e2**2))
        I1 = Gamma1
        I2 = Gamma2
        
        Ltot = L1 + L2 - I1 - I2
        Psi = -k * ( s * L2 + (1+s) * L1) 
        rt_a20 = Ltot / (self.beta2 + self.beta1 * np.sqrt(alpha_res)) 

        I1 /= rt_a20
        I2 /= rt_a20
        Psi  /= rt_a20
        Ltot /= rt_a20
        L1byL2res = self.beta1 * np.sqrt(alpha_res) / self.beta2 
        denom = L1byL2res + s * (1 + L1byL2res)
        amd = (-Psi/k)  * (1+L1byL2res) / denom - Ltot 
        sigma1 = theta1 / k
        sigma2 = theta2 / k
        x1 = np.sqrt(2 * I1) * np.cos(sigma1)
        y1 = np.sqrt(2 * I1) * np.sin(sigma1)
        x2 = np.sqrt(2 * I2) * np.cos(sigma2) 
        y2 = np.sqrt(2 * I2) * np.sin(sigma2)

        #dscale = 1 / action_scale**2
        #tscale = dscale**(1.5)
        #scales = {'action':action_scale,'distance':dscale, 'time':tscale}

        return np.array((y1,y2,x1,x2,amd))
    
    def find_equilibrium(self,guess,dissipation=False,tolerance=1e-9,max_iter=10):
        """
        Use Newton's method to locate an equilibrium solution of 
        the equations of motion. 
    
        By default, an equilibrium of the dissipation-free equations
        is sought. In this case, the AMD value of the equilibrium 
        solution will be equal to the value of the initially supplied 
        guess of dynamical variables. If dissipative terms are included
        then the equilibrium will depend on the parameters K1 and K2
        as well as tau_alpha.
    
        Arguments
        ---------
        guess : ndarray
            Initial guess for the dynamical variables at the equilibrium.
        dissipation : bool, optional
            Whether dissipative terms are considered in the equations of
            motion. Default is False.
        tolerance : float, optional
            Tolerance for root-finding such that solution satisfies :math:`|f(z)|` < tolerance
            Default value is 1E-9.
        max_iter : int, optional
            Maximum number of Newton's method iterations.
            Default is 10. If maximum is reached, result will be 
            returned with a warning.
        include_dissipation : bool, optional
            Include dissipative terms in the equations of motion.
            Default is False
    
        Returns
        -------
        zeq : ndarray
            Equilibrium value of dynamical variables.
    
        """
        if dissipation:
            return self._find_dissipative_equilibrium(guess,tolerance,max_iter)
        else:
            return self._find_conservative_equilibrium(guess,tolerance,max_iter)

    def _find_dissipative_equilibrium(self,guess,tolerance=1e-9,max_iter=10):
        y = guess
        fun = self.flow
        jac = self.flow_jac
        f = fun(y)
        J = jac(y)
        it=0
        # Newton method iteration
        while np.linalg.norm(f)>tolerance and it < max_iter:
            dy = lsq_linear(J,-f).x
            y = y + dy
            f = fun(y)
            J = jac(y)
            it+=1
            if it==max_iter:
                warn("Newton's method failed to converge before the maximum of {} iterations were completed. Try re-running with a higher value of 'max_iter' or choose a better initial guess for the equilibrium configuration.".format(max_iter))
        return y
    def _find_conservative_equilibrium(self,guess,tolerance=1e-9,max_iter=10):
        y = guess
        fun = self.H_flow
        jac = self.H_flow_jac
        f = fun(y)[:-1]
        J = jac(y)[:-1,:-1]
        it=0
        # Newton method iteration
        while np.linalg.norm(f)>tolerance and it < max_iter:
            dy = lsq_linear(J,-f).x
            y[:-1] = y[:-1] + dy
            f = fun(y)[:-1]
            J = jac(y)[:-1,:-1]
            it+=1
            if it==max_iter:
                raise RuntimeError("Max iterations reached!")
        return y

    def dyvars_to_rebound_simulation(self,z,Q=0,pomega1=0,osculating_correction = True,include_dissipation = False,**kwargs):
        r"""
        Convert dynamical variables

        .. math::
            z = (y_1,y_2,x_1,x_2,{\cal D}

        to a Rebound simulation.

        Arguments
        ---------
        z : ndarray
            Dynamical variables
        pomega1 : float, optional
            Planet 1's longitude of periapse. 
            Default is 0
        Q : float, optional
            Angle variable Q = (lambda2-lambda1) / k. 
            Default is Q=0. Resonant periodic orbits
            are 2pi-periodic in Q.
        osculating_correction : bool, optional
            If True, apply correction from the 'averaged' variables employed by
            the resonance model to osculating variables used in the un-averaged
            three-body problem. Correction is done to first order in planet masses
            and assumes that the system is near an equilibrium configuration of 
            the resonance model.
            Default is True.
        include_dissipation : bool, optional
            Include dissipative effects through 
            reboundx's external forces.
            Default is False

        Keyword Arguments
        -----------------
        inc : float, default = 0.0
            Inclination of planets' orbits.

        Returns
        -------
        tuple :
            Returns a tuple. The first item
            of the tuple is a rebound simulation. 
            The second item is a reboundx.Extras object
            if 'include_dissipation' is True, otherwise
            the second item is None.
        """
        mean_orbels = self.dyvars_to_orbital_elements(z)
        if osculating_correction:
            zosc = self.mean_to_osculating_dyvars(Q,z)
            orbels = self.dyvars_to_orbital_elements(zosc)
        else:
            orbels = mean_orbels
        j,k = self.j, self.k
        keys=['a1','e1','theta1','a2','e2','theta2']
        a1,e1,theta1,a2,e2,theta2 = [orbels[key] for key in keys]

        syn_angle = self.k * Q
        pomega2 = np.mod( pomega1 -  (theta2 - theta1) / k, 2*np.pi)
        M1 = np.mod( (theta1 - j*syn_angle) / k ,2*np.pi )
        l1 = np.mod(M1 + pomega1,2*np.pi)
        l2 = np.mod( syn_angle + l1,2*np.pi)
        inc = kwargs.pop('inc',0.0)

        sim = rb.Simulation()
        sim.add(m=1)
        for i in range(1,3):
            els={'a':orbels['a{}'.format(i)],'e':orbels['e{}'.format(i)],'inc':inc}
            els['l']=[l1,l2][i-1]
            els['omega']=[pomega1,pomega2][i-1]
            els['Omega']=0.
            m = [self.m1,self.m2][i-1]
            add_canonical_heliocentric_elements_particle(m,els,sim)
        sim.move_to_com()
        
        if include_dissipation:
            ps = sim.particles
            rebx = reboundx.Extras(sim)
            mod = rebx.load_operator("modify_orbits_direct")
            rebx.add_operator(mod)
            mod.params["p"] = self.p
            timescales = self.timescales
            ps[1].params["tau_a"]=-1*timescales['tau_m1'] 
            ps[2].params["tau_a"]=-1*timescales['tau_m2'] 
            ps[1].params["tau_e"]=-1*timescales['tau_e1'] 
            ps[2].params["tau_e"]=-1*timescales['tau_e2'] 
            return sim,rebx
        else:
            return sim
    def dyvars_from_rebound_simulation(self,sim,iIn=1,iOut=2, osculating_correction = False):
        r"""
        Convert dynamical variables

        .. math:
            z = (\sigma_1,\sigma_2,I_1,I_2,{\cal C}

        to a Rebound simulation.

        Arguments
        ---------
        sim : rebound.Simulation
            Simulation object to 
        iIn : int, optional
            Integer index of inner planet to use
            for computing dynamical variables.
            Default is 1
        iOut : int, optional 
            Integer index of outer planet to use
            for computing dynamical variables.
            Default is 2
        osculating_correction : boole, optional
            If True, correct the orbital elements
            of the rebound simulation by transforming
            to the mean variables of the resonance model.
        Returns
        ---------
        z : ndarray
            Dynamical variables
        """
        els=dict()
        orbits = get_canonical_heliocentric_orbits(sim)
        o1 = orbits[iIn-1]
        o2 = orbits[iOut-1]
        for i,o in enumerate([o1,o2]):
            els['a{}'.format(i+1)] = o.a
            els['e{}'.format(i+1)] = o.e
            els['pomega{}'.format(i+1)] = o.pomega
            els['l{}'.format(i+1)] = o.l
        theta = self.j * els['l2'] - (self.j-self.k) * els['l1']
        for i in range(1,3):
            els['theta{}'.format(i)] = np.mod(theta-self.k * els['pomega{}'.format(i)],2*np.pi)
        if osculating_correction:
            warnings.warn("Osculating corrections are currently not implemented.")
        return self.orbital_elements_to_dyvars(els)

    def dyvars_to_Poincare_actions(self,dyvars):
        """
        Convert a set of dynamical variables to Poincare actions

        Arguments
        ---------
        dyvars : ndarray
          Dynamical variables to convert

        Returns
        -------
        actions : dict
          Poincare actions stored as dictionary entries.
        """
        return self._funcs['actions'](dyvars,*self.extra_args)

    def mean_to_osculating_dyvars(self,Q,z,N = 256):
        r"""
        Apply perturbation theory to transfrom from the phase space coordiantes of the
        averaged model to osculating phase space coordintates of the full phase space.
        Assumes that the transformation is being applied at the fixed point of the
        averaged model. Details are described in the appendix of 
        `Hadden & Payne (2020) <https://ui.adsabs.harvard.edu/abs/2020AJ....160..106H/abstract>`_

        Arguemnts
        ---------
        Q : float or ndarry
            Value(s) of angle :math:`Q=(lambda2-lambda1)/k` at which to apply transformation.
            Equilibrium points of the averaged model correspond to orbits that are
            periodic in Q in the full phase space.
        z : ndarray
            Dynamical variables of the averaged model, math::`z =
            (y_1,y_2,x_1,x_2,{\cal D})`

        N : int, optional
            Number of Q points to evaluate functions at when performing Fourier
            transformation. Default is N=256

        Returns
        -------
        zosc : ndarray, (5,) or (M,5)
            The osculating values of the dynamical varaibles for the
            input Q values. The dimension of the returned variables
            is set by the dimension of the input 'Q'. If Q is a
            float, then z is an array of length 5. If Q is an
            array of length M then zosc is an (M,5) array.
        """
        omega_syn = self.omega_syn(z)
        OmegaMtrx = getOmegaMatrix(2)
        Omega_del2H = self.H_flow_jac(z)[:-1,:-1]
        vals,S = np.linalg.eig(Omega_del2H)
        S = np.transpose([S.T[i] for i in (0,2,1,3)])
        s1 = (S.T @ OmegaMtrx @ S)[0,2]
        s2 = (S.T @ OmegaMtrx @ S)[1,3]
        S.T[0]*=1/np.sqrt(s1)
        S.T[2]*=1/np.sqrt(s1)
        S.T[1]*=1/np.sqrt(s2)
        S.T[3]*=1/np.sqrt(s2)
        Sinv = np.linalg.inv(S)

        Qarr = np.atleast_1d(Q)
        dchi = np.zeros((4,len(Qarr)),dtype=np.complex128)

        # Fill arrays for FFT
        Qs = np.linspace(0,2*np.pi,N)
        X = np.zeros((N,4),dtype=np.complex128)

        gradHkep = -1 * OmegaMtrx @ self.H_kep_flow(z)[:-1]
        domega_syn_dz = self.grad_omega_syn(z)
        domega_syn_dw = S.T @ domega_syn_dz

        for i,q in enumerate(Qs):
            gradHosc = self.grad_H_pert_full(q,z)[:-1] + gradHkep
            X[i] = S.T @ (gradHosc)
            X[i] -= self.H_pert_osc(q,z) * domega_syn_dw/omega_syn

        omegas = +1*np.imag(np.diag(Sinv @ Omega_del2H @ S))[:2]
        for I in range(2):
            A = np.fft.fft(X[:,I])
            freqs = np.fft.fftshift(np.fft.fftfreq(N)*N)
            amps = np.fft.fftshift(A)/N
            for l in range(1,N//2 - 1):
                sig = -1j * self.k * amps[N//2+l] * np.exp(1j * freqs[N//2+l] * Qarr)  / (-l*omega_syn - self.k * omegas[I])
                sig +=-1j * self.k * amps[N//2-l] * np.exp(1j * freqs[N//2-l] * Qarr)  / (l*omega_syn - self.k * omegas[I])
                dchi[I] += sig

        dchi[2] = -1j * np.conjugate(dchi[0])
        dchi[3] = -1j * np.conjugate(dchi[1])

        # Get AMD correction
        j,k = self.j,self.k
        L20 = self.beta2
        L10 = self.beta1 * np.sqrt(self.alpha)
        f = L10/L20
        s = (j-k)/k
        prefactor = (1+f) / (f*(1+s) + s) / k
        dAMD = prefactor * np.array([self.H_pert_osc(q,z) for q in Qarr]) / omega_syn 

        dz = np.transpose(-1 * (S @ OmegaMtrx @ dchi).T)
        result = np.transpose(z + np.vstack((dz,dAMD)).T)
        result = np.real(result) # trim small imaginary parts cause by numerical errors
        if result.shape[1] == 1:
            return result.reshape(-1)
        return result

    def integrate_initial_conditions(self,dyvars0,times,dissipation=False):
        r"""
        Integrate initial conditions and calculate dynamical
        variables and orbital elements at specified times.

        Integration is done with the scipy.integrate.odeint
        method. 

        Arguments
        ---------
        dyvars0 : ndarray
            Initial conditions in the form of an array
            of dynamical variables.
        times : ndarray
            Times at which to calculate output.

        Returns
        -------
        soln : dict
            Dictionary containing both the dynamical variables
            and the orbital elements of the integrated solution.
            The dictionary keys are
            - 'times': Times at which solutin is computed.
            - 'dynamical_variables': ndarray of dynamical at each of the times in 'times'.
            - 'orbital_elements': dict containing arrays of the various orbital elements.
        """
        if dissipation:
            f = lambda y,t: self.flow(y)
            Df = lambda y,t: self.flow_jac(y)
        else:
            f = lambda y,t: self.H_flow(y)
            Df = lambda y,t: self.H_flow_jac(y)
        soln_dyvars = odeint(
                f,
                dyvars0,
                times,
                Dfun = Df
        )
        el_dict0=self.dyvars_to_orbital_elements(dyvars0)
        els_dict = {key:np.zeros(len(soln_dyvars)) for key in el_dict0};
        els_dict['times'] = times
        for i,sol in enumerate(soln_dyvars):
            els = self.dyvars_to_orbital_elements(sol)
            for key,val in els.items():
                els_dict[key][i]=val
        return {'times':times,'dynamical_variables':soln_dyvars,'orbital_elements':els_dict}

#######################
####### Spatial #######
#######################
def _get_spatial_compiled_theano_functions(N_QUAD_PTS):
    # Planet masses: m1,m2
    m1,m2 = T.dscalars(2)
    mstar = 1
    mu1  = m1 * mstar / (mstar  + m1) 
    mu2  = m2 * mstar / (mstar  + m2) 
    Mstar1 = mstar + m1
    Mstar2 = mstar + m2
    beta1 = mu1 * T.sqrt(Mstar1/mstar) / (mu1 + mu2)
    beta2 = mu2 * T.sqrt(Mstar2/mstar) / (mu1 + mu2)
    j,k = T.lscalars('jk')
    s = (j-k) / k

    # Angle variable for averaging over
    psi = T.dvector()

    # Quadrature weights
    quad_weights = T.dvector('w')

    # Dynamical variables:
    Ndof = 3
    Nconst = 1
    dyvars = T.vector()
    y1, y2, y_inc, x1, x2, x_inc, amd  = [dyvars[i] for i in range(2*Ndof + Nconst)]
    
    a20 = T.constant(1.)
    a10 = ((j-k)/j)**(2/3) * (Mstar1 / Mstar2)**(1/3)
    L10 = beta1 * T.sqrt(a10)
    L20 = beta2 * T.sqrt(a20)
    Ltot = L10 + L20
    f = L10/L20
    L2res = (Ltot + amd) / (1+f)
    Psi = -k * (s * L2res + (1+s) * f * L2res) 
    ###
    # actions
    ###
    I1 = 0.5 * (x1*x1 + y1*y1)
    I2 = 0.5 * (x2*x2 + y2*y2)
    Phi = 0.5 * (x_inc*x_inc + y_inc*y_inc)
    L1 = -s*Ltot - Psi/k - s * (I1 + I2 + Phi)
    L2 = (1+s)*Ltot + Psi/k + (1+s) * (I1 + I2 + Phi)

    # Set lambda2=0
    l2 = T.constant(0.)
    l1 = -1 * k * psi 
    theta_res = (1+s) * l2 - s * l1
    cos_theta_res = T.cos(theta_res)
    sin_theta_res = T.sin(theta_res)
    
    kappa1 = x1 * cos_theta_res + y1 * sin_theta_res
    eta1   = y1 * cos_theta_res - x1 * sin_theta_res
    
    kappa2 = x2 * cos_theta_res + y2 * sin_theta_res
    eta2   = y2 * cos_theta_res - x2 * sin_theta_res
    
    sigma = x_inc * cos_theta_res + y_inc * sin_theta_res
    rho   = y_inc * cos_theta_res - x_inc * sin_theta_res
    # y = (sigma-i*rho)/sqrt(2)
    #   = sqrt(Phi) * exp[i (Omega1+Omega2) / 2]
    # Malige+ 2002,  Eqs 20 and 21
    r2byr1 = (L2 - L1 - I2 + I1) / Ltot
    sigma1 = rho * T.sqrt( 1 + r2byr1) / T.sqrt(2)
    sigma2 = -rho * T.sqrt( 1 - r2byr1) / T.sqrt(2)
    rho1 = -sigma * T.sqrt( 1 + r2byr1) / T.sqrt(2)
    rho2 = sigma * T.sqrt( 1 - r2byr1) / T.sqrt(2)

    Xre1 = kappa1 / T.sqrt(L1)
    Xim1 = -eta1 / T.sqrt(L1)
    Yre1 = 0.5 * sigma1 / T.sqrt(L1)
    Yim1 = -0.5 * rho1 / T.sqrt(L1)

    Xre2 = kappa2 / T.sqrt(L2)
    Xim2 = -eta2 / T.sqrt(L2)
    Yre2 = 0.5 * sigma2 / T.sqrt(L2)
    Yim2 = -0.5 * rho2 / T.sqrt(L2)

    absX1_sq = 2 * I1 / L1
    absX2_sq = 2 * I2 / L2
    X_to_z1 = T.sqrt(1 - absX1_sq / 4 )
    X_to_z2 = T.sqrt(1 - absX2_sq / 4 )
    Y_to_zeta1 =  1 / T.sqrt(1 - absX1_sq/2) 
    Y_to_zeta2 =  1 / T.sqrt(1 - absX2_sq/2) 

    a1 = (L1 / beta1 )**2 
    k1 = Xre1 * X_to_z1
    h1 = Xim1 * X_to_z1
    q1 = Yre1 * Y_to_zeta1
    p1 = Yim1 * Y_to_zeta1
    e1 = T.sqrt( absX1_sq ) * X_to_z1
    inc1 = 2*T.arcsin(T.sqrt(p1*p1+q1*q1))

    a2 = (L2 / beta2 )**2 
    k2 = Xre2 * X_to_z2
    h2 = Xim2 * X_to_z2
    q2 = Yre2 * Y_to_zeta2
    p2 = Yim2 * Y_to_zeta2
    e2 = T.sqrt( absX2_sq ) * X_to_z2
    inc2 = 2*T.arcsin(T.sqrt(p2*p2+q2*q2))

    beta1p = T.sqrt(Mstar1) * beta1
    beta2p = T.sqrt(Mstar2) * beta2
    Hkep = -0.5 * beta1p / a1 - 0.5 * beta2p / a2

    Hdir,Hind = calc_Hint_components_spatial(
            a1,a2,l1,l2,h1,k1,h2,k2,p1,q1,p2,q2,Mstar1,Mstar2
    )
    eps = m1*m2/ (mu1 + mu2) / T.sqrt(mstar)
    Hpert = eps * (Hdir + Hind/mstar)
    Hpert_av = Hpert.dot(quad_weights)
    Htot = Hkep + Hpert_av

    #####################################################
    # Set parameters for compiling functions with Theano
    #####################################################
    
    # Get numerical quadrature nodes and weights
    nodes,weights = np.polynomial.legendre.leggauss(N_QUAD_PTS)
    
    # Rescale for integration interval from [-1,1] to [-pi,pi]
    nodes = nodes * np.pi
    weights = weights * 0.5
    
    # 'givens' will fix some parameters of Theano functions compiled below
    givens = [(psi,nodes),(quad_weights,weights)]

    # 'ins' will set the inputs of Theano functions compiled below
    #   Note: 'extra_ins' will be passed as values of object attributes
    #   of the 'ResonanceEquations' class 'defined below
    extra_ins = [m1,m2,j,k]
    ins = [dyvars] + extra_ins

    Stilde = Phi * (L2-I2 - L1 + I1) / (Ltot)
    Q1 = 0.5 * (Phi+Stilde)
    Q2 = 0.5 * (Phi-Stilde)
    inc1 = T.arccos(1-Q1/(L1-I1))
    inc2 = T.arccos(1-Q2/(L2-I2))

    orbels = [a1,e1,inc1,k*T.arctan2(y1,x1),a2,e2,inc2,k*T.arctan2(y2,x2),T.arctan2(y_inc,x_inc)]
    orbels_dict = dict(zip(
            ['a1','e1','inc1','theta1','a2','e2','inc2','theta2','phi'],
            orbels
        )
    )

    actions = [L1,L2,I1,I2,Q1,Q2]
    actions_dict = dict(
            zip(
                ['L1','L2','Gamma1','Gamma2','Q1','Q2'],
                actions
                )
            )

    #  Conservative flow
    gradHtot = T.grad(Htot,wrt=dyvars)
    gradHpert = T.grad(Hpert_av,wrt=dyvars)
    gradHkep = T.grad(Hkep,wrt=dyvars)

    hessHtot = theano.gradient.hessian(Htot,wrt=dyvars)
    hessHpert = theano.gradient.hessian(Hpert_av,wrt=dyvars)
    hessHkep = theano.gradient.hessian(Hkep,wrt=dyvars)

    Jtens = T.as_tensor(np.pad(getOmegaMatrix(Ndof),(0,Nconst),'constant'))
    H_flow_vec = Jtens.dot(gradHtot)
    Hpert_flow_vec = Jtens.dot(gradHpert)
    Hkep_flow_vec = Jtens.dot(gradHkep)

    H_flow_jac = Jtens.dot(hessHtot)
    Hpert_flow_jac = Jtens.dot(hessHpert)
    Hkep_flow_jac = Jtens.dot(hessHkep)

    ##########################
    # Compile Theano functions
    ##########################
    func_dict={
     # Hamiltonians
     'H':Htot,
     #'Hpert':Hpert_av,
     #'Hkep':Hkep,
     ## Hamiltonian flows
     'H_flow':H_flow_vec,
     #'Hpert_flow':Hpert_flow_vec,
     #'Hkep_flow':Hkep_flow_vec,
     ## Hamiltonian flow Jacobians
     'H_flow_jac':H_flow_jac,
     #'Hpert_flow_jac':Hpert_flow_jac,
     #'Hkep_flow_jac':Hkep_flow_jac,
     ## Extras
     'orbital_elements':orbels_dict,
     'actions':actions_dict
    }
    compiled_func_dict=dict()
    with tqdm(func_dict.items()) as t:
        for key,val in t:
            t.set_description("Compiling '{}'".format(key))
            if key == 'timescales':
                inputs = extra_ins
            else:
                inputs = ins 
            cf = theano.function(
                inputs=inputs,
                outputs=val,
                givens=givens,
                on_unused_input='ignore'
            )
            compiled_func_dict[key]=cf
    return compiled_func_dict


class SpatialResonanceEquations():

    """
    A class for equations describing the dynamics of a pair of 
    planets in/near a mean motion resonance.

    Attributes
    ----------
    j : int
        Together with k specifies j:j-k resonance
    
    k : int
        Order of resonance.
    
    alpha : float
        Semi-major axis ratio a_1/a_2

    eps : float
        Mass parameter m1*m2 / (mu1+mu2)

    m1 : float
        Inner planet mass

    m2 : float
        Outer planet mass

    """
    def __init__(self,j,k, n_quad_pts = 40, m1 = 1e-5 , m2 = 1e-5):
        self.j = j
        self.k = k
        self.m1 = m1
        self.m2 = m2
        self.mstar = 1
        self.n_quad_pts = n_quad_pts
        self._funcs = _get_spatial_compiled_theano_functions(n_quad_pts)

    @property
    def extra_args(self):
        return [self.m1,self.m2,self.j,self.k]
    @property
    def mu1(self):
        return self.m1 / (1 + self.m1)
    @property
    def mu2(self):
        return self.m2 / (1 + self.m2)
    @property
    def beta1(self):
        return self.mu1 * np.sqrt(1+self.m1) / (self.mu1 + self.mu2)
    @property
    def beta2(self):
        return self.mu2 * np.sqrt(1+self.m2) / (self.mu1 + self.mu2)
    @property
    def eps(self):
        return self.m1 * self.m2 / (self.mu1 + self.mu2)
    @property
    def alpha(self):
        alpha0 = ((self.j-self.k)/self.j)**(2/3)
        return alpha0 * ((1 + self.m1) / (1+self.m2))**(1/3)
    def H(self,z):
        """
        Calculate the value of the Hamiltonian.

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            The value of the Hamiltonian evaluated at z.
        """
        return self._funcs['H'](z,*self.extra_args)

    def H_kep(self,z):
        """
        Calculate the value of the Keplerian component of the Hamiltonian.

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            The value of the Keplerian part of the Hamiltonian evaluated at z.
        """
        return self._funcs['Hkep'](z,*self.extra_args)

    def H_pert(self,z):
        r"""
        Calculate the value of the perturbation component of the Hamiltonian.

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        float : 
            The value of the perturbation part of the Hamiltonian evaluated at z.
        """
        return self._funcs['Hpert'](z,*self.extra_args)

    def H_flow(self,z):
        r"""
        Calculate flow induced by the Hamiltonian

        .. math::
            \dot{z} = \Omega \cdot \nabla_{z}H(z)

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        ndarray : 
            Flow vector
        """
        return self._funcs['H_flow'](z,*self.extra_args)


    def H_flow_jac(self,z):
        r"""
        Calculate the Jacobian of the flow induced by the Hamiltonian
      
        .. math::
             \Omega \cdot \Delta H(z)

        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        ndarray : 
            Jacobian matrix
        """
        return self._funcs['H_flow_jac'](z,*self.extra_args)

    def dyvars_to_orbital_elements(self,z):
        r"""
        Convert dynamical variables
        .. math:
            z = (y_1,y_2,y_inc,x_1,x_2,y_inc,{\cal D})
        to orbital elements
        .. math:
            (a_1,e_1,inc_1,\theta_1,a_2,e_2,inc_2,\theta_2)
        """

        return self._funcs['orbital_elements'](z,*self.extra_args)

    def dyvars_to_Poincare_actions(self,z):
        r"""
        Convert dynamical variables
        .. math:
            z = (y1,y2,yinc,x1,x2,xinc,amd) 
        to canonical Poincare action variables
        .. math:
            (L1,L2,Gamma1,Gamma2,Q1,Q2)
        
        Arguments
        ---------
        z : ndarray
            Dynamical variables

        Returns
        -------
        dict : 
            Dictionary of canonical actions with strings as keys.
        """

        return self._funcs['actions'](z,*self.extra_args)

    def dyvars_from_rebound_simulation(self,sim,iIn=1,iOut=2,osculating_correction=False,full_output=False):
        r"""
        Convert rebound simulation to dynamical variables.

        Arguments
        ---------
        sim : rebound.Simulation
            Simulation object
        iIn : int, optional 
            Index of inner planet in simulation particles
            list. Default is 1.
        iOut : int, optional 
            Index of outer planet in simulation particles
            list. Default is 2.
        osculating_correction : boole, optional
            If True, correct the orbital elements
            of the rebound simulation by transforming
            to the mean variables of the resonance model.
            Default is False.
        full_output : boole, optional
            Return a dictionary containing the re-scaling 
            factors from time and distance units of the
            input simulation to the time and distance units
            used by SpatialResonanceEquations. 
        Returns
        -------
        z : ndarray
            Dynamical variables for resonance equations of motion.
        scales : dict, optional
            Returned if 'full_output' is True. 
            'scales' contains conversion factors from 
            simulation to resonance equation units such that:
               distance_sim = scales['distance'] * distance_eqs
               time_sim = scales['time'] * time_eqs
        """
        if osculating_correction:
            warnings.warn("Osculating corrections are currently not implemented.")
        align_simulation(sim)
        star = sim.particles[0]
        p1 = sim.particles[iIn]
        p2 = sim.particles[iOut]
        
        self.m1 = p1.m/star.m
        self.m2 = p2.m/star.m
        orbits = get_canonical_heliocentric_orbits(sim)
        o1 = orbits[iIn-1]
        o2 = orbits[iOut-1]
        
        L1 = self.beta1 * np.sqrt(o1.a)
        L2 = self.beta2 * np.sqrt(o2.a)
        Gamma1 = L1 * (1 - np.sqrt(1-o1.e**2))
        Gamma2 = L2 * (1 - np.sqrt(1-o2.e**2))
        Q1 = (L1 - Gamma1) * (1 - np.cos(o1.inc))
        Q2 = (L2 - Gamma2) * (1 - np.cos(o2.inc))
        
        k = self.k
        j = self.j
        Mstar = self.mstar
        m1 = self.m1
        m2 = self.m2
        alpha_res = self.alpha
        s = (j - k) / k
        
        Ltot =  L1 + L2 - Gamma1 - Gamma2 - Q1 - Q2
        factor =  self.beta2 + self.beta1 * np.sqrt(alpha_res)
        action_scale = (factor/Ltot)
        L1 *= action_scale
        L2 *= action_scale
        Gamma1 *= action_scale
        Gamma2 *= action_scale
        Q1 *= action_scale
        Q2 *= action_scale
        Ltot *= action_scale
        dscale = 1 / action_scale**2
        tscale = dscale**(1.5)
        scales = {'action':action_scale,'distance':dscale, 'time':tscale}
        
        I1 = Gamma1
        I2 = Gamma2
        Phi = Q1 + Q2
        Psi = -k * ( s * L2 + (1+s) * L1) 

        s1 = (1 + s) * o2.l - s * o1.l - o1.pomega
        s2 = (1 + s) * o2.l - s * o1.l - o2.pomega
        phi = (1 + s) * o2.l - s * o1.l - 0.5 * (o1.Omega + o2.Omega)
        
        y1 =  np.sqrt(2*I1)*np.sin(s1)
        x1 =  np.sqrt(2*I1)*np.cos(s1)
        y2 =  np.sqrt(2*I2)*np.sin(s2)
        x2 =  np.sqrt(2*I2)*np.cos(s2)
        y_inc =  np.sqrt(2*Phi)*np.sin(phi)
        x_inc =  np.sqrt(2*Phi)*np.cos(phi)

        L1byL2res = self.beta1 * np.sqrt(alpha_res) / self.beta2 
        denom = L1byL2res + s * (1 + L1byL2res)
        amd = (-Psi/k)  * (1+L1byL2res) / denom - Ltot 

        if full_output:
            return np.array([y1,y2,y_inc,x1,x2,x_inc,amd]),scales
        else:
            return np.array([y1,y2,y_inc,x1,x2,x_inc,amd])

    def mean_to_osculating_dyvars(self,psi,z,N = 256):
        r"""
        Apply perturbation theory to transfrom from the phase space coordiantes of the 
        averaged model to osculating phase space coordintates of the full phase space.

        Assumes that the transformation is being applied at the fixed point of the 
        averaged model.

        Arguemnts
        ---------
        Q : float or ndarry
            Value(s) of angle (lambda2-lambda1)/k at which to apply transformation.
            Equilibrium points of the averaged model correspond to orbits that are 
            periodic in Q in the full phase space.
        z : ndarray
            Dynamical variables of the averaged model, :math:`z =
            (y_1,y_2,y_I,x_1,x_2,x_I,\mathcal{D})`
        N : int, optional
            Number of Q points to evaluate functions at when performing Fourier
            transformation. Default is N=256

        Returns
        -------
        zosc : ndarray, (7,) or (M,7) 
            The osculating values of the dynamical varaibles for the
            input Q values. The dimension of the returned variables
            is set by the dimension of the input 'Q'. If Q is a 
            float, then z is an array of length 5. If Q is an 
            array of length M then zosc is an (M,5) array.
        """
        warnings.warn("Correction from mean to osculating variables is not currently implemented!")
        return z

    def dyvars_to_rebound_simulation(self,z,psi=0,Omega=0,osculating_correction = True,include_dissipation = False,**kwargs):
        r"""
        Convert dynamical variables to a Rebound simulation.

        Arguments
        ---------
        z : ndarray
            Dynamical variables
        Omega : float, optional
            Value of cyclic coordinate Omega   
        psi : float, optional
            Angle variable psi = (lambda1-lambda2) / k. 
            Default is psi=0. Resonant periodic orbits
            are 2pi-periodic in Q.
        include_dissipation : bool, optional
            Include dissipative effects through 
            reboundx's external forces.
            Default is False

        Keyword Arguments
        -----------------
        Mstar : float
            Default=1.0
            Stellar mass.
        inc : float, default = 0.0
            Inclination of planets' orbits.
        period1 : float
            Default = 1.0
            Orbital period of inner planet
        units : tuple
            Default = ('AU','Msun','days')
            Units of rebound simulation.

        Returns
        -------
        tuple :
            Returns a tuple. The first item
            of the tuple is a rebound simulation. 
            The second item is a reboundx.Extras object
            if 'include_dissipation' is True, otherwise
            the second item is None.
        """
        mean_orbels = self.dyvars_to_orbital_elements(z)
        if osculating_correction:
            zosc = self.mean_to_osculating_dyvars(psi,z)
            orbels = self.dyvars_to_orbital_elements(zosc)
        else:
            orbels = mean_orbels
        j,k = self.j, self.k
        s = (j-k) / k
        sim = rb.Simulation()
        Mstar = self.mstar
        mpl1 = self.m1 
        mpl2 = self.m2 
        orbels['l1'] = orbels['phi'] + k * (1 + s) * psi + Omega
        orbels['l2'] = orbels['phi'] + k * s * psi + Omega
        orbels['pomega1'] = orbels['phi'] + Omega - orbels['theta1']/k
        orbels['pomega2'] = orbels['phi'] + Omega - orbels['theta2']/k
        orbels['Omega1'] = np.pi/2 - Omega
        orbels['Omega2'] = -np.pi/2 - Omega
        for i in range(1,3):
            orbels['omega{}'.format(i)] = orbels['pomega{}'.format(i)] - orbels['Omega{}'.format(i)]
        sim.add(m=Mstar)
        for i,mpl in enumerate([mpl1,mpl2]):
            elements = {key:orbels[key + str(i+1)] for key in ['a','e','inc','l','omega','Omega']}
            add_canonical_heliocentric_elements_particle(mpl,elements,sim)
        sim.move_to_com()
        rebx = None
        if include_dissipation:
            pass
        return sim

    def integrate_initial_conditions(self,dyvars0,times):
        """
        Integrate initial conditions and calculate dynamical
        variables and orbital elements at specified times.

        Integration is done with the scipy.integrate.odeint
        method. 

        Arguments
        ---------
        dyvars0 : ndarray
            Initial conditions in the form of an array
            of dynamical variables.
        times : ndarray
            Times at which to calculate output.
        """
        f = lambda y,t: self.H_flow(y)
        Df = lambda y,t: self.H_flow_jac(y)
        soln_dyvars = odeint(
                f,
                dyvars0,
                times,
                Dfun = Df
        )
        el_dict0=self.dyvars_to_orbital_elements(dyvars0)
        els_dict = {key:np.zeros(len(soln_dyvars)) for key in el_dict0};
        els_dict['times'] = times
        for i,sol in enumerate(soln_dyvars):
            els = self.dyvars_to_orbital_elements(sol)
            for key,val in els.items():
                els_dict[key][i]=val
        return {'times':times,'dynamical_variables':soln_dyvars,'orbital_elements':els_dict}
